/** Copyright (C) 1993 McGill University.
    Copyright (C) 1993 McCAT System Group.
    Copyright (C) 1993 Bhama Sridharan
    Copyright (C) 1993 Justiani
    Copyright (C) 1993 Clark Verbrugge

This program is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 1, or (at your option) any later
   version.

This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 675
   Mass Ave, Cambridge, MA 02139, USA. **/

/******************************************************************
*                                                                       *
*       Routines for dumping C-code from AST                            *
*                                                                       *
*******************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>             /* LJH added to be able to comp filename with
                                   __FAKE_IF */
#include "tree.h"
#include "flags.h"
#include "c-tree.h"
#include "c-parse.h"
#include "bind-lev-str.h"
#include "general.h"
#include "General.h"
#include "messages.h"
#include "new-alloc.h"
/********************************************************************/
/* JUSTIANI's include                                               */
/********************************************************************/
#include "analysis.h"
#include "dam.h"
#include "c-dump.h"
#include "dam-dump.h"
#include "alias.h"
#include "heap.h"
/********************************************************************/
/* Includes needed for dump-hook                                    */
/********************************************************************/
#include "dh.h"

/* PO: temporary hack for using this new c-dump with the older McCAT (pre
   atomic, shared and basic). */

#ifndef TREE_ATOMIC
#define TREE_ATOMIC(NODE) 0
#endif

#ifndef TREE_SHARED
#define TREE_SHARED(NODE) 0
#endif

#ifndef TREE_BASIC
#define TREE_BASIC(NODE) TREE_LOCAL(NODE)
#endif

/* PO: end of temporary hack. */

/* EARTH dump */
typedef enum e_End_type
{
  EMPTY, SEQ, PAR, STRICT_SEQ
}     End_type;

static char rcsid[] = "$Id: c-dump.c,v 1.58 1996/11/27 04:37:02 ghiya Exp $";

IMPORT FILE *al_file;
IMPORT int flag_alias;          /* the flag for the alias to be processed */
IMPORT int flag_connect_dump;
IMPORT int Alias_dump;          /* the flag to print the result of alias
                                   process */
IMPORT int Alias_indirect_dump; /* the flag to print the alias info only for
                                   indirect refs in the program */
IMPORT int flag_print_fields; /* the flag to print field sets for heap ptrs */
IMPORT int flag_connect_rw_dump; /* the flag to print heap r/w sets */
IMPORT int dump_read_write_sets; /* the flag to print read/write sets  */
IMPORT int dump_EARTH_comments; /* the flag for printing EARTH comments */
IMPORT int EARTH_dump;          /* Flag for dumping EARTH-C */
IMPORT int flag_dump_explicit_types;    /* Discard typedef names when dumping */
/* Unused */

/* the prototype for all functions    */
/* -------------------------------    */

/*--------------------------------*/
/* Global variables */
/*--------------------------------*/


/* Rakesh : Dec 10/96: flag to indicate we are dumping a func prototype */
LOCAL int dump_func_prototype;
LOCAL int dump_typedefs;
/* Current structure for which fields are being dumped : Rakesh Dec 13/96*/
LOCAL tree current_struct_node;
static dumphook *dh_lists[DH_MAX];
static int dh_hook(int wl, tree node);
EXPORT tree dump_cur_func;      /* Rakesh */
LOCAL void new_spit_type_i(tree, unsigned long, tree, int, int);
LOCAL tree get_nth_type_constr(tree, int);
LOCAL int get_tc_length_from_type(tree type);
LOCAL unsigned long build_tc_from_type(tree type);
/* Rakesh: Nov 26/96 : check if the parent of an addr_expr is array ref */
LOCAL int parent_is_array_ref(tree node, tree stmt, int side);
LOCAL int height_difference(tree addr_type, tree op_type);

/** empty string for proper indentation                                 **/
char  indent_str[INDENT_size];

/** This is a common file pointer, which points to either the c dump    **/
/** file or the structure file                                          **/
FILE *c_file;

/** name of the c-dump-file and the c-structures dump file              **/
FILE *c_dump_file;
FILE *c_struct_dump_file;
/* All the typedefs are dumped here: Rakesh Dec 16/96 */
FILE *c_typedef_dump_file; 
/* All the variable declarations are dumped here: Rakesh Dec 16/96 */
FILE *c_var_dump_file; 
/* HTML stuff to show the info as frames: Rakesh Jan 13/96 */
FILE *frames_dump_file; 
/* File to which the current flow info being displayed will be copied */
FILE *info_dump_file; 
FILE *note_dump_file; 
FILE *menu_dump_file; 
FILE *connect_indirect_dump_file; 
FILE *connect_dump_file;
FILE *fields_dump_file;
FILE *heap_rw_dump_file;
FILE *alias_indirect_dump_file;
FILE *alias_dump_file;
/** the name of the structure file to be included at the top of the     **/
/** c_dump file                                                         **/
char *struct_name;

char *comment_string;
void  get_comment_string();

extern struct binding_level *global_binding_level;
extern char *tree_code_name[];
extern tree unknown_arr_size_node;

/** has the name of the input c file. this has been set in toplev.c     **/
extern char *ast_c_dump_base_name;
static char *c_base_name; /* the name of file excluding the path name */
char *full_path_name; /* absolute path name to the directory from where
                       * cc1 is executed: obtained from the getcwd command*/

/** flag to determine whether or not c-dump is to be done               **/
extern int flag_c_dump;
extern int flag_earth, flag_earth_syntax;
extern int flag_compact_dump;   /* Unused */

/** list of structures to be output to the '___.cstr.h' file            **/
/** at the end.                                                         **/
EXPORT tree struct_list = NULL_TREE;

/** list of typedefs found.                                             **/
LOCAL tree typedef_list = NULL_TREE;

static int current_column = 0;
static int save_column;
static int flag_fold_and_align = 0;
static const int right_margin = 78; 

/** current_indentation                                                 **/
LOCAL int curr_indent = 0;

/** the flag whether spit string "End of list" or not                   **/
int   f_spit_end = 1;

/* Prototypes */

static int basic_type(tree decl);
static int is_basic_term(tree expr);
void  dec_indent(void);
void  inc_indent(void);
static int dh_hook(int wl, tree node);
static int file_or_string(int set);
static void open_c_files(void);
static void spit_all_structures(void);
static void spit_arg(tree arg, tree stmt);
static void spit_bit_field_size(tree decl);
static void
spit_check_operands(tree node, char *str, int mod_flag,
  int no_of_ops, tree stmt, int side);
static void spit_end_comp_stmt(End_type end_type);
static void spit_expr(tree node, int mod_flag, tree stmt, int side);
static void spit_func_prototype(tree decl, int indent_flag);
static void spit_initializations(tree decl);
static void
spit_operand(tree node, char *str, int mod_flag, tree stmt,
  int side);
static void
spit_operands(tree node, char *str, int mod_flag, tree stmt,
  int side);
static void spit_params(tree params);
static void spit_type_qual(tree type);
static void spit_storage_class(tree node);
static void spit_start_comp_stmt(End_type end_type);
static void spit_stmt(tree node, End_type end_type);
static void spit_tagless_struct(tree node);
static void spit_var(tree decl, int indent_flag);
static void spit_vars(tree decl);

static int is_fake_if(tree if_stmt);
EXPORT void spit_decl(tree decl, int indent_flag, int spit_base_type_flag);
LOCAL void spit_all_frames();
LOCAL void spit_all_menus();
LOCAL void spit_all_typedefs(tree ast_list);
LOCAL void spit_all_globals(tree ast_list);
static void spit_array_size(tree array_type);
static void spit_type_cast(tree type_cast);
static void spit_n_spaces(int n);
static int get_nb_columns_after_last_newline(char *str);

/* SSA Numbering dump additions. By C. Lapkowski */
static void spit_SSA_Number(tree decl, tree stmt, int side);
extern int get_SSA_number(tree decl, tree stmt, int side, int *ext_number);
#define SSA_RHS 1
#define SSA_LHS 2
/* End Addition */
/* From  heap_analysis.c: Rakesh: Jan 8/97 */
IMPORT INDEX_INFO heap_get_index(tree node, tree func);


/** called to determine the names of the c-dump file and the structures **/
/** file and to open them for writing                                   **/
LOCAL void
open_c_files(void)
{
  int   dump_base_name_length, length;
  register char *dumpname, *benchname;
  char *lastslash;
  char *base_name;              /* get just name, not the path */
  char *temp;                   /* assist in stripping off the path */

  full_path_name == NULL;
  dump_base_name_length = 64;
  while (full_path_name == NULL) 
  {
    if ((full_path_name = getcwd((char *)NULL, dump_base_name_length)) == NULL)
    { 
      dump_base_name_length+=64;
    } 
    else
     break;
   }
  temp = c_base_name = ast_c_dump_base_name;
  while (*temp != '\0')
  {
    if (*temp == '/')
    {
      c_base_name = ++temp;
      length = 0;
      while (*temp != '\0')
      {
        if (*temp == '/')
        {
          break;
        }
        length++;
        temp++;
      }
      break;
    }
    else
      temp++;
  }
  benchname = (char *) Mmalloc(length + 20);
  strncpy(benchname, c_base_name, length);
  full_path_name = (char *) Mmalloc(length + 40);
  sprintf(full_path_name, "~ghiya/flow/%s", benchname);
  fprintf(stderr, "%s\n", ast_c_dump_base_name);

  dump_base_name_length = strlen(ast_c_dump_base_name);

  /*** the size of the dump base name is dump_base_name_length + the
  suffix required-- either .cstr.h or .dump.c etc.  **/

  dumpname = (char *) Mmalloc(dump_base_name_length + 20);

  /* strip off the leading path, and have c_base_name point to this name */
  /* Rakesh Dec 13/96 : used for interfile html refs */
  temp = c_base_name = ast_c_dump_base_name;
  while (*temp != '\0')
  {
    if (*temp == '/')
    {
      c_base_name = ++temp;
    }
    else
      temp++;
  }


  if (EARTH_dump)
    sprintf(dumpname, "%1.*s.earth.c",
      dump_base_name_length - 2, ast_c_dump_base_name);
  else if (flag_earth)
    sprintf(dumpname, "%1.*s.threaded.c",
      dump_base_name_length - 2, ast_c_dump_base_name);
  else
    sprintf(dumpname, "%1.*s.c.html",
      dump_base_name_length - 2, ast_c_dump_base_name);

  fprintf(stderr, "Opening file '%s'\n", dumpname);
  c_dump_file = fopen(dumpname, "w");

  /*--------- MAR -----------*/
  al_file = c_dump_file;
  /*-------------------------*/
  if (c_dump_file == 0)
    pfatal_with_name(dumpname);

  /**MYDEBUG**/
  fprintf(stderr, "c-dump-base-name = %s\n", dumpname);

  /* Rakesh : Dec 13/96 */
  sprintf(dumpname, "%s.cstr.html", ast_c_dump_base_name);

  /* strip off the leading path, and have base_name point to this name */
  temp = base_name = dumpname;
  while (*temp != '\0')
  {
    if (*temp == '/')
      base_name = ++temp;
    else
      temp++;
  }

  struct_name = base_name;

  /**MYDEBUG**/
  fprintf(stderr, "struct_file name = %s\n", dumpname);

  c_struct_dump_file = fopen(dumpname, "w");

  if (c_struct_dump_file == 0)
    pfatal_with_name(dumpname);

  /* Rakesh : Dec 13/96 */
  sprintf(dumpname, "%s.typedefs.html", ast_c_dump_base_name);
  fprintf(stderr, "typedef_dump_file name = %s\n", dumpname);
  c_typedef_dump_file = fopen(dumpname, "w");
  if (c_typedef_dump_file == 0)
    pfatal_with_name(dumpname);

  /* Rakesh : Dec 13/96 */
  sprintf(dumpname, "%s.vars.html", ast_c_dump_base_name);
  fprintf(stderr, "var_dump_file name = %s\n", dumpname);
  c_var_dump_file = fopen(dumpname, "w");
  if (c_var_dump_file == 0)
    pfatal_with_name(dumpname);

  /* Rakesh : Jan 13/97 */
  sprintf(dumpname, "%s.frames.html", ast_c_dump_base_name);
  fprintf(stderr, "frames_dump_file name = %s\n", dumpname);
  frames_dump_file = fopen(dumpname, "w");
  if (frames_dump_file == 0)
    pfatal_with_name(dumpname);
  /* Rakesh : Jan 13/97 */
  sprintf(dumpname, "%s.info.html", ast_c_dump_base_name);
  fprintf(stderr, "info_dump_file name = %s\n", dumpname);
  info_dump_file = fopen(dumpname, "w");
  if (info_dump_file == 0)
    pfatal_with_name(dumpname);
  sprintf(dumpname, "%s.note.html", ast_c_dump_base_name);
  fprintf(stderr, "note_dump_file name = %s\n", dumpname);
  note_dump_file = fopen(dumpname, "w");
  if (note_dump_file == 0)
    pfatal_with_name(dumpname);
  /* Rakesh : Jan 13/97 */
  sprintf(dumpname, "%s.menu.html", ast_c_dump_base_name);
  fprintf(stderr, "menu_dump_file name = %s\n", dumpname);
  menu_dump_file = fopen(dumpname, "w");
  if (menu_dump_file == 0)
    pfatal_with_name(dumpname);
  /* Rakesh : Jan 16/97 */
  if (flag_connect_indirect_dump)
  {
    sprintf(dumpname, "%s.connect.i.html", ast_c_dump_base_name);
    fprintf(stderr, "connect_indirect_dump_file name = %s\n", dumpname);
    connect_indirect_dump_file = fopen(dumpname, "w");
    if (connect_indirect_dump_file == 0)
      pfatal_with_name(dumpname);
  }
  if (flag_connect_dump)
  {
    sprintf(dumpname, "%s.connect.html", ast_c_dump_base_name);
    fprintf(stderr, "connect_dump_file name = %s\n", dumpname);
    connect_dump_file = fopen(dumpname, "w");
    if (connect_dump_file == 0)
      pfatal_with_name(dumpname);
  }
  if (flag_print_fields)
  {
    sprintf(dumpname, "%s.fields.html", ast_c_dump_base_name);
    fprintf(stderr, "fields_dump_file name = %s\n", dumpname);
    fields_dump_file = fopen(dumpname, "w");
    if (fields_dump_file == 0)
      pfatal_with_name(dumpname);
  }
  if (flag_connect_rw_dump)
  {
    sprintf(dumpname, "%s.heap.rw.html", ast_c_dump_base_name);
    fprintf(stderr, "heap_rw_dump_file name = %s\n", dumpname);
    heap_rw_dump_file = fopen(dumpname, "w");
    if (heap_rw_dump_file == 0)
      pfatal_with_name(dumpname);
  }
  /* Rakesh : Jan 19/97 */
  if (Alias_indirect_dump)
  {
    sprintf(dumpname, "%s.alias.i.html", ast_c_dump_base_name);
    fprintf(stderr, "alias_indirect_dump_file name = %s\n", dumpname);
    alias_indirect_dump_file = fopen(dumpname, "w");
    if (alias_indirect_dump_file == 0)
      pfatal_with_name(dumpname);
  }
  /* Rakesh : Jan 20/97 */
  if (Alias_dump)
  {
    sprintf(dumpname, "%s.alias.html", ast_c_dump_base_name);
    fprintf(stderr, "alias_dump_file name = %s\n", dumpname);
    alias_dump_file = fopen(dumpname, "w");
    if (alias_dump_file == 0)
      pfatal_with_name(dumpname);
  }
}

/** main routine to dump c code. Goes through the list of all top       **/
/** level function and variable declarations and spits out the c code   **/
/** keeps track of all the structures found and spits them in the       **/
/** structures file.                                                    **/
EXPORT void
spit_all_c_code(tree list)
{
  tree  ast_node;
  tree  ast_list;
  tree  func;
  tree  temp;

  if (!(flag_c_dump))
    return;

  open_c_files();

  if (mydebug)
    fprintf(stderr, "STARTING TO C_DUMP FUNCTIONS\n");

  c_file = c_dump_file;

  fprintf(c_var_dump_file, "<HTML>\n");
  fprintf(c_var_dump_file, "<PRE>\n");
  if (connect_indirect_dump_file)
  {
    fprintf(connect_indirect_dump_file, "<HTML>\n");
    fprintf(connect_indirect_dump_file, "<PRE>\n");
    fprintf(connect_indirect_dump_file, "INFOHEADER=connect_indirect\n");
  }
  if (connect_dump_file)
  {
    fprintf(connect_dump_file, "<HTML>\n");
    fprintf(connect_dump_file, "<PRE>\n");
    fprintf(connect_dump_file, "INFOHEADER=simple_connect\n");
  }
  if (heap_rw_dump_file)
  {
    fprintf(heap_rw_dump_file, "<HTML>\n");
    fprintf(heap_rw_dump_file, "<PRE>\n");
    fprintf(heap_rw_dump_file, "INFOHEADER=heap_rw_sets\n");
    
    fprintf(heap_rw_dump_file, "CSTRFILE=%s/%s.cstr.html\n", full_path_name, 
            c_base_name);
  }
  if (fields_dump_file)
  {
    fprintf(fields_dump_file, "<HTML>\n");
    fprintf(fields_dump_file, "<PRE>\n");
    fprintf(fields_dump_file, "INFOHEADER=connect_fields\n");
    
    fprintf(fields_dump_file, "CSTRFILE=%s/%s.cstr.html\n", full_path_name, 
            c_base_name);
  }
  if (alias_indirect_dump_file)
  {
    fprintf(alias_indirect_dump_file, "<HTML>\n");
    fprintf(alias_indirect_dump_file, "<PRE>\n");
    fprintf(alias_indirect_dump_file, "INFOHEADER=alias_indirect\n");
  }
  if (alias_dump_file)
  {
    fprintf(alias_dump_file, "<HTML>\n");
    fprintf(alias_dump_file, "<PRE>\n");
    fprintf(alias_dump_file, "INFOHEADER=simple_alias\n");
  }
  fprintf(c_file, "<HTML>\n");
  fprintf(c_file, "<BODY bgcolor=white LINK=black>\n");
  /* default target is "info" for all links in the source file */
  fprintf(c_file,"<BASE HREF=\"http://www-acaps.cs.mcgill.ca/\" ");
  /* fprintf(c_file, "<BASE  "); */
  fprintf(c_file, "TARGET = \"info\">\n");
  fprintf(c_file, "<TABLE>\n"); /* to get colors in links */
  fprintf(c_file, "<TD>\n");
  fprintf(c_file, "<PRE>\n");
  fprintf(c_file, "<A HREF=\"%s/%s.cstr.html\"><FONT color=blue>struct defs</FONT></a>\n", full_path_name, c_base_name);
  fprintf(c_file, "\n<A HREF=\"%s/%s.typedefs.html\"><FONT color=blue>typedefs</FONT></a>\n", full_path_name, c_base_name);
  fprintf(c_file, "\n<A HREF=\"%s/%s.vars.html\"><FONT color=blue>global vars</FONT></a>\n", full_path_name, c_base_name);
  fprintf(c_file, "\n");

  dump_func_prototype = 0;
  dump_typedefs = 0;
  ast_list = list;
  dh_hook(DH_BEFORE, list);

  /* Rakesh : Dec 16/96 : all typedefs are now dumped in a separate 
     file which can be accessed using an html link from the main file */
  spit_all_typedefs(list);
  spit_all_globals(list);
  ast_list = list;
  while (ast_list)
  {
    if (TREE_CODE(AST_DECL_PTR(ast_list)) == FUNCTION_DECL)
    {
      dump_cur_func = AST_DECL_PTR(ast_list);
      if (TREE_CODE(TREE_TYPE(AST_DECL_PTR(ast_list))) != FUNCTION_TYPE)
        fprintf(stderr, "FUNCTION_DECL does not have FUNCTION_TYPE\n");

      /* Rakesh : Nov 17/96: Don't dump psuedo-codes from prototypes.c */
      if (IS_PSEUDO_FUNCTION(AST_DECL_PTR(ast_list)))
      {
        ast_list = TREE_CHAIN(ast_list);
        continue;
      }
      if (TREE_TYPE(AST_DECL_PTR(ast_list)) != NULL_TREE)
        TYPE_ARG_LIST(TREE_TYPE(AST_DECL_PTR(ast_list)))
          = DECL_ARGUMENTS(AST_DECL_PTR(ast_list));
      if (AST_DECL_BODY(ast_list) != NULL_TREE)
      {
        dh_hook(DH_BEFOREFUNC, ast_list);

        spit_newline();
	/* non-called function bodies are dumped in italics : Rakesh Jan 27/97*/
	if (flag_callgraph && (!IS_IN_CALL_GRAPH(dump_cur_func)))
	{
	  spit_html_string("<EM>");
	}
        spit_storage_class(AST_DECL_PTR(ast_list));
        spit_decl(AST_DECL_PTR(ast_list), NO_INDENT, 1);
        spit_newline();
        spit_stmt(AST_DECL_BODY(ast_list), EMPTY);
	if (flag_callgraph && (!IS_IN_CALL_GRAPH(dump_cur_func)))
	{
	  spit_html_string("</EM>");
	}
        dh_hook(DH_AFTERFUNC, ast_list);
      }
      else
      {
	if (flag_callgraph && (!IS_IN_CALL_GRAPH(AST_DECL_PTR(ast_list))))
	{
	  spit_html_string("<EM>");
	}
        spit_func_prototype(AST_DECL_PTR(ast_list), NO_INDENT);
	if (flag_callgraph && (!IS_IN_CALL_GRAPH(AST_DECL_PTR(ast_list))))
	{
	  spit_html_string("</EM>");
	}
      }
    }
#if 0
    else
    {
      if (TREE_CODE(AST_DECL_PTR(ast_list)) != TYPE_DECL)
      { /* Don't dump global vars from prototypes.c file */
        if (!IS_PSEUDO_FUNCTION(AST_DECL_PTR(ast_list)))
        {
	  /* dump the variable declarations in a separate file,
	   * accessible from the main c-dump file via an html link */
	  c_file = c_var_dump_file;
	  if (first_global == 0)
	  {
	   spit_html_string("<H2> Global Variables </h2>");
	   first_global = 1;
	  }
          spit_decl(AST_DECL_PTR(ast_list), INDENT, 1);
          spit_string(";\n");
	  /* reset the c_file to the main c-dump file */
	  c_file = c_dump_file;
        }
      }
    }
#endif
    ast_list = TREE_CHAIN(ast_list);
    curr_indent = 0;
  }

  dh_hook(DH_AFTER, list);

  fprintf(c_file, "\n</PRE>\n");
  fprintf(c_file, "</TD>\n");
  fprintf(c_file, "</TABLE>\n");
  fprintf(c_file, "</BODY>\n");
  fprintf(c_file, "</HTML>\n");
  fclose(c_file);

  /****
    if (mydebug)
    ***/
  fprintf(stderr, "STARTING TO DUMP STRUCTURES \n");

  c_file = c_struct_dump_file;
  curr_indent = 0;
  spit_all_structures();
  fclose(c_file);
  fprintf(c_var_dump_file, "\n</PRE>\n");
  fprintf(c_var_dump_file, "</HTML>\n");
  fclose(c_var_dump_file);
  if (connect_indirect_dump_file)
  {
    fprintf(connect_indirect_dump_file, "\n</PRE>\n");
    fprintf(connect_indirect_dump_file, "</HTML>\n");
    fclose(connect_indirect_dump_file);
  }
  if (connect_dump_file)
  {
    fprintf(connect_dump_file, "\n</PRE>\n");
    fprintf(connect_dump_file, "</HTML>\n");
    fclose(connect_dump_file);
  }
  if (heap_rw_dump_file)
  {
    fprintf(heap_rw_dump_file, "\n</PRE>\n");
    fprintf(heap_rw_dump_file, "</HTML>\n");
    fclose(heap_rw_dump_file);
  }
  if (fields_dump_file)
  {
    fprintf(fields_dump_file, "\n</PRE>\n");
    fprintf(fields_dump_file, "</HTML>\n");
    fclose(fields_dump_file);
  }
  if (alias_indirect_dump_file)
  {
    fprintf(alias_indirect_dump_file, "\n</PRE>\n");
    fprintf(alias_indirect_dump_file, "</HTML>\n");
    fclose(alias_indirect_dump_file);
  }
  if (alias_dump_file)
  {
    fprintf(alias_dump_file, "\n</PRE>\n");
    fprintf(alias_dump_file, "</HTML>\n");
    fclose(alias_dump_file);
  }
  if (mydebug)
    fprintf(stderr, "FINISHED  DUMP STRUCTURES \n");
  spit_all_frames();
  spit_all_menus();
  fprintf(stderr, "FINISHED C DUMP \n");
}

/** spits out a list of variables-- calls spit_var for each variable    **/
LOCAL void
spit_vars(tree decl)
{
  tree  save = decl;
  tree  last_decl = NULL_TREE;
  int   nb_decls = 0;

  c_file = c_var_dump_file;
  while (decl)
  {
    if (TREE_CODE(decl) != FUNCTION_DECL)
    {
      /* Check whether the current declaration has the same base name as the
         previous one.  If so, append the declaration to the current line
         (using a comma), rather than starting a new declaration on a
         different line.  Note that qualifiers and storage classes are
         syntactically associated with the base name, so these need to be the
         same as well.
      
      This does not handle the case
      
      int i, a[10];
      
      as the types are not the same. */
      if (last_decl != NULL_TREE
        && TREE_CODE(last_decl) == VAR_DECL && TREE_CODE(decl) == VAR_DECL
        && DECL_TYPE_NAME(decl) == DECL_TYPE_NAME(last_decl)
        && TREE_REGDECL(decl) == TREE_REGDECL(last_decl)
        && TREE_STATIC(decl) == TREE_STATIC(last_decl)
        && TREE_READONLY(decl) == TREE_READONLY(last_decl)
        && TREE_VOLATILE(decl) == TREE_VOLATILE(last_decl)
        && TREE_SHARED(decl) == TREE_SHARED(last_decl)
        && TREE_TYPE(decl) == TREE_TYPE(last_decl))
      {
        spit_string(", ");
        save_column = 2 + strlen(indent_str);
        flag_fold_and_align = 1;
        spit_decl(decl, NO_INDENT, 0);
        flag_fold_and_align = 0;
        nb_decls += 1;
      }
      else
      {
        if (last_decl)
          spit_string(";\n");

        if (nb_decls == 0) /* Rakesh : first var being dumped */
	{
	  spit_html_target_string(c_file, dump_cur_func, NULL);
          spit_html_string("\n <H2> Local vars for the Function\n ");
          spit_html_link_string(c_file, dump_cur_func, 
           IDENTIFIER_POINTER(DECL_NAME(dump_cur_func)), 
           c_base_name, NULL, "source");

          spit_html_string("</H2>\n");
	}
        spit_decl(decl, INDENT, 1);
        nb_decls += 1;
      }
    }
    if (TREE_CODE(decl) != FUNCTION_DECL)
      last_decl = decl;

    decl = TREE_CHAIN(decl);
  }
  if (last_decl)
    spit_string(";\n");

  if (nb_decls > 0)
   spit_newline();
  c_file = c_dump_file;
  /* If declarations were spit, output an empty line to separate the
     declaration section fron the body. */
  if (nb_decls > 0)
  {
    spit_string(indent_str);
    spit_html_link_string(c_file, dump_cur_func, "local vars", 
                          c_base_name, ".vars.html", NULL);
    spit_newline();
    spit_newline();
  }
}

/* PO: spits the type qualifiers for a type node. */

LOCAL void
spit_type_qual(tree node)
{
  if (TREE_READONLY(node))
    spit_string("const ");

  if (TREE_VOLATILE(node))
    spit_string("volatile ");

  /* PO: The EARTH-specific type qualifiers are dumped here. */
  if (flag_earth_syntax && (EARTH_dump || dump_EARTH_comments))
  {
    if (dump_EARTH_comments)
      spit_string("/* ");

    if (TREE_LOCAL(node))
      spit_string("local ");

    if (TREE_SHARED(node))
      spit_string("shared ");
    /* Add other EARTH qualifiers here, for example
    
    spit_string("earth_qual1 ");
    
    the space before the closing quote is important. */

    if (dump_EARTH_comments)
      spit_string("*/ ");
  }
}

/** spits out storage classes like extern and static                **/
LOCAL void
spit_storage_class(tree node)
{
  if (TREE_REPLICATED(node) && TREE_PUBLIC(node))
  {
    if (dump_EARTH_comments)
      spit_string("/* replicated */ ");
    else if (EARTH_dump)
      spit_string("replicated ");
  }
  if (dump_EARTH_comments && !TREE_REPLICATED(node) && TREE_PUBLIC(node)
    && TREE_CODE(node) != FUNCTION_DECL)
    spit_string("/* global */ ");

  if (flag_earth_syntax && TREE_BASIC(node))
  {
    if (dump_EARTH_comments)
      spit_string("/* basic */ ");
    else if (EARTH_dump)
      spit_string("basic ");
  }
  if (flag_earth_syntax && TREE_ATOMIC(node))
  {
    if (dump_EARTH_comments)
      spit_string("/* atomic */ ");
    else if (EARTH_dump)
      spit_string("atomic ");
  }
  if (!flag_earth_syntax && TREE_EXTERNAL(node))
    spit_string("extern ");

  if (TREE_REGDECL(node))
    spit_string("register ");

  if ((!TREE_PUBLIC(node)) && (TREE_STATIC(node)))
    spit_string("static ");
}

static void
spit_bit_field_size(tree decl)
{
  if ((DECL_SIZE(decl) != NULL) && (TREE_LITERAL(DECL_SIZE(decl))))
    if (TREE_INT_CST_LOW(DECL_SIZE(decl)) == 1)
    {
      char *temp = (char *) Mmalloc(1);
      sprintf(temp, " : %d", DECL_SIZE_UNIT(decl));
      spit_string(temp);
    }
}

static void
spit_initializations(tree decl)
{
  if (TREE_CODE(decl) == VAR_DECL)
    if ((DECL_INITIAL(decl) != NULL_TREE) &&
      (DECL_INITIAL(decl) != error_mark_node))
    {
      spit_string(" = ");
      spit_expr(DECL_INITIAL(decl), 0, NULL, 0);
    }
}

/** returns 1 if the node is in typedef list-- used to determine        **/
/** if a structure has a typedef name                                   **/
LOCAL int
in_typedef_list(node)
  tree  node;
{
  tree  temp;

  temp = typedef_list;
  while (temp)
  {
    if (node == TREE_VALUE(temp))
    {
      spit_name(TREE_PURPOSE(temp));
      spit_string(" ");
      return (1);
    }
    temp = TREE_CHAIN(temp);
  }
  return (0);
}


/** spits out the type of a variable/function. If typedef flag          **/
/** is 1, the typedef list is not checked to see if the type has        **/
/** an user defined typedef name, otherwise it is.                      **/
/** If the variable is of array type, this routine returns the string   **/
/** to be output after the array name-- e.g. [12][13].                  **/


char *
spit_type(tree node, int typedef_flag, int precedence_flag)
{
  char *newstr, *str_ptr, *bracket;
  tree  values, fields;

  if (node == NULL_TREE)
    return (NULL_STRING);

  if (!typedef_flag)
    if (in_typedef_list(node))
      return (NULL_STRING);

  if ((TYPE_NAME(node)) && (TREE_CODE(TYPE_NAME(node)) == TYPE_DECL) && (!(typedef_flag)))
  {
    /* spit_string(IDENTIFIER_POINTER(DECL_NAME(TYPE_NAME(node)))); */
    /* Rakesh: Dec 17/96: Set a link to the typedef definition */
    spit_html_link_string(c_file, TYPE_NAME(node), 
     IDENTIFIER_POINTER(DECL_NAME(TYPE_NAME(node))), 
     c_base_name, ".typedefs.html", NULL); 
    spit_string(" ");
    return NULL_STRING;
  }
  switch (TREE_CODE(node))
  {
  case VOID_TYPE:
    spit_string("void ");
    return (NULL_STRING);
  case INTEGER_TYPE:
    if (TYPE_NAME(node) != NULL)
    {
      if (TREE_CODE(TYPE_NAME(node)) == IDENTIFIER_NODE)
        spit_string(IDENTIFIER_POINTER(TYPE_NAME(node)));
      else if (TREE_CODE(TYPE_NAME(node)) == TYPE_DECL && DECL_NAME(TYPE_NAME(node)) != NULL)
        spit_string(IDENTIFIER_POINTER(DECL_NAME(TYPE_NAME(node)))); 
      spit_string(" ");
    }
    else
    {
      /** This else is not necessary at all; TYPE NAME
      can never be NULL in INTEGER and FLOAT types **/
      /** (I think!!) **/
      if (TREE_UNSIGNED(node))
        spit_string("unsigned ");
      switch (TYPE_PRECISION(node))
      {
      case INT_TYPE_SIZE:
        spit_string("int ");
        return (NULL_STRING);
        /****
    INT_TYPE_SIZE and LONG_TYPE_SIZE have the same
    size here.
    case LONG_TYPE_SIZE:
        spit_string("long ");
        return(NULL_STRING);
        ***/
      case SHORT_TYPE_SIZE:
        spit_string("short ");
        return (NULL_STRING);
      case LONG_LONG_TYPE_SIZE:
        spit_string("long long ");
        return (NULL_STRING);
      case CHAR_TYPE_SIZE:
        spit_string("char ");
        return (NULL_STRING);
      default:
        fprintf(stderr, "error in spit_type\n");
        return (NULL_STRING);
      }
    }
    break;
  case REAL_TYPE:
    if (TYPE_NAME(node) != NULL)
    {
      if (TREE_CODE(TYPE_NAME(node)) == IDENTIFIER_NODE)
        spit_string(IDENTIFIER_POINTER(TYPE_NAME(node)));
      else if (TREE_CODE(TYPE_NAME(node)) == TYPE_DECL && DECL_NAME(TYPE_NAME(node)) != NULL)
        spit_string(IDENTIFIER_POINTER(DECL_NAME(TYPE_NAME(node)))); 
      spit_string(" ");
    }
    else
    {
      /** This else is not necessary at all; TYPE NAME
      can never be NULL in integer and FLOAT types **/
      switch (TYPE_PRECISION(node))
      {
      case FLOAT_TYPE_SIZE:
        spit_string("float ");
        return (NULL_STRING);
      case DOUBLE_TYPE_SIZE:
        spit_string("double ");
        return (NULL_STRING);
        /***************
    case LONG_DOUBLE_TYPE_SIZE:
        spit_string("long double  ");
        return(NULL_STRING);
        ************/
      default:
        fprintf(stderr, "error in spit_type\n");
        return (NULL_STRING);
      }
    }
    break;
  case POINTER_TYPE:
    /* handle precedence of * and [] properly */
    /* Added by Christopher Lapkowski */
    /* Proper handling of pointer to functions should be done too */
    /* however; this is not a simple task. A hack can be done but */
    /* it would not be nice */
    /* Bhama Do the same for function pointers too */

    if ((!((TYPE_NAME(TREE_TYPE(node))) &&
          (TREE_CODE(TYPE_NAME(TREE_TYPE(node))) == TYPE_DECL))) &&
      ((TREE_CODE(TREE_TYPE(node)) == ARRAY_TYPE) ||
        (TREE_CODE(TREE_TYPE(node)) == FUNCTION_TYPE)))
    {
      str_ptr = spit_type(TREE_TYPE(node), typedef_flag, TRUE);
      spit_string("( * ");
    }
    else
    {
      str_ptr = spit_type(TREE_TYPE(node), typedef_flag, FALSE);
      spit_string(" * ");
    }
    return (str_ptr);

  case ARRAY_TYPE:
    {
      int   size, num;

      str_ptr = spit_type(TREE_TYPE(node), typedef_flag, FALSE);

      /*** newstr should be large enough to   **/
      /** hold the rest of the dimensions     **/
      /** plus this dimension          **/
      /** 32 = size of a number + 2 for       **/
      /** brackets **/
      num = strlen(str_ptr) + 1 + 20;
      newstr = (char *) (Mmalloc(num));

      if (precedence_flag)
        bracket = ")";
      else
        bracket = "";

      if (TYPE_DOMAIN(node))
      {
        size = get_int_cst(TYPE_MAX_VALUE(TYPE_DOMAIN(node)));
        if (size == -1)
        {
          if (str_ptr)
            sprintf(newstr, "%s[]%s", bracket, str_ptr);
          else
            sprintf(newstr, "%s[]", bracket);
        }
        else
        {
          if (str_ptr)
            sprintf(newstr, "%s[%d]%s", bracket, size + 1, str_ptr);
          else
            sprintf(newstr, "%s[%d]", bracket, size + 1);
        }
      }
      else
      {
        if (str_ptr)
          sprintf(newstr, "%s[]%s", bracket, str_ptr);
        else
          sprintf(newstr, "%s[]", bracket);
      }
      return (newstr);
    }

  case RECORD_TYPE:
    if (TYPE_NAME(node) && (TREE_CODE(TYPE_NAME(node)) ==
        IDENTIFIER_NODE))
    {
      spit_string("struct ");
      spit_string(IDENTIFIER_POINTER(TYPE_NAME(node)));
      spit_string(" ");
      fields = TYPE_FIELDS(node);
      return (NULL_STRING);
    }
    else
    {
      spit_tagless_struct(node);
    }
    return (NULL_STRING);

  case UNION_TYPE:
    if (TYPE_NAME(node) && (TREE_CODE(TYPE_NAME(node)) ==
        IDENTIFIER_NODE))
    {
      spit_string("union ");
      spit_string(IDENTIFIER_POINTER(TYPE_NAME(node)));
      spit_string(" ");
      fields = TYPE_FIELDS(node);
      return (NULL_STRING);
    }
    else
    {
      spit_tagless_struct(node);
    }

    return (NULL_STRING);

  case FUNCTION_TYPE:
    {
      int   num;

      str_ptr = spit_type(TREE_TYPE(node), typedef_flag, FALSE);

      if (precedence_flag)
        bracket = ")";
      else
        bracket = "";

      num = strlen(str_ptr) + 5;
      newstr = (char *) (Mmalloc(num));
      if (str_ptr)
        sprintf(newstr, "%s%s()", bracket, str_ptr);
      else
        sprintf(newstr, "%s()", bracket);
      return (newstr);
    }

  case ENUMERAL_TYPE:
    if (TYPE_NAME(node) && (TREE_CODE(TYPE_NAME(node)) ==
        IDENTIFIER_NODE))
    {
      spit_string("enum ");
      spit_string(IDENTIFIER_POINTER(TYPE_NAME(node)));

      spit_string(" ");
    }
    else
    {
      /*** this is to take care of the case   ***/
      /*** typedef enum { .... } DATA;                ***/
      spit_string("enum ");
      spit_string(" ");
      spit_string(" { ");
      values = TYPE_VALUES(node);
      while (values)
      {
        spit_string(IDENTIFIER_POINTER(TREE_PURPOSE(values)));
        if (TREE_VALUE(values))
        {
          spit_string(" = ");
          spit_cst(TREE_VALUE(values));
        }
        values = TREE_CHAIN(values);
        if (values)
          spit_string(",");
      }
      spit_string(" } ");
    }

    return (NULL_STRING);
  case TYPE_DECL:
    spit_name(node);
    spit_string(" ");
    return (NULL_STRING);
  default:
    fprintf(stderr, "Wrong opcode in spit_type\n");
    fputs(tree_code_name[(int) TREE_CODE(node)], stderr);
    exit(NULL_STRING);

  }
}

/** inserts in the global struct_list a new structure that has been     **/
/** found. The list of structures with names , e.g. struct mm { .. }    **/
/** are pointed to by a field in the binding level structure.           **/
/** This list is added to struct_list, plus any new nameless structures **/
/** like struct { ... } are also added.                                 **/
EXPORT
insert_in_struct_list(node)
  tree  node;
{
  int   found;
  tree  temp, new;

  if (struct_list == NULL_TREE)
    struct_list = build_tree_list(NULL_TREE, node);
  else
  {
    temp = struct_list;
    found = 0;
    while (temp)
    {
      if (TREE_VALUE(temp) == node)
        found = 1;
      temp = TREE_CHAIN(temp);
    }
    if (!found)
    {
      new = build_tree_list(NULL_TREE, node);
      struct_list = chainon(struct_list, new);
    }
  }
}

LOCAL void
spit_all_typedefs(tree ast_list)
{
  dump_typedefs = 1;
  c_file = c_typedef_dump_file;
  fprintf(c_file, "<HTML>\n");
  fprintf(c_file, "<PRE>\n");
  while (ast_list)
  {
    if (TREE_CODE(AST_DECL_PTR(ast_list)) == TYPE_DECL)
    {
      spit_decl(AST_DECL_PTR(ast_list), INDENT, 1);
      spit_string(";\n");
    }
    ast_list = TREE_CHAIN(ast_list);
  }
  dump_typedefs = 0;
  fprintf(c_file, "\n</PRE>\n");
  fprintf(c_file, "</HTML>\n");
  fclose(c_file);
}

LOCAL void
spit_all_globals(tree ast_list)
{
  int first_global;
  first_global = 0;
  c_file = c_var_dump_file;
  while (ast_list)
  {
    if ((TREE_CODE(AST_DECL_PTR(ast_list)) != TYPE_DECL) &&
        (TREE_CODE(AST_DECL_PTR(ast_list)) != FUNCTION_DECL)) 
    { /* Don't dump global vars from prototypes.c file */
      if (!IS_PSEUDO_FUNCTION(AST_DECL_PTR(ast_list)))
      {
        if (first_global == 0)
        {
           spit_html_string("<H2> Global Variables </h2>");
	   first_global = 1;
         }
         spit_decl(AST_DECL_PTR(ast_list), INDENT, 1);
         spit_string(";\n");
      }
    }
  ast_list = TREE_CHAIN(ast_list);
  }
  c_file = c_dump_file;
}
/** spits all structures into the structures file. Called by    **/
/** spit_all_c_code , in the end.                               **/
LOCAL void
spit_all_structures(void)
{
  tree  fields;
  tree  temp;
  tree  values;
  tree  node;

  fprintf(c_file, "<HTML>\n");
  fprintf(c_file, "<PRE>\n");
  temp = struct_list;
  current_struct_node = NULL;

  while (temp)
  {
    node = TREE_VALUE(temp);

    if (!TREE_EXTERNAL(node))
    {
      switch (TREE_CODE(node))
      {
      case RECORD_TYPE:
      case UNION_TYPE:
        if (TREE_CODE(node) == RECORD_TYPE)
          spit_string("struct ");
        else
          spit_string("union ");

        if (TYPE_NAME(node))
	{
          /* spit_string(IDENTIFIER_POINTER(TYPE_NAME(node)));*/
          /* Pass TYPE_NAME(node) as in heap_collect_fields.c, the
           * field vars point to TYPE_NAME(node) instead of node itself:
           * so to maintain correct labels pass the former : Rakesh Dec 13/96 */
          spit_html_target_string(c_file, TYPE_NAME(node), 
                                  IDENTIFIER_POINTER(TYPE_NAME(node)));
          current_struct_node = TYPE_NAME(node);
	}
        fields = TYPE_FIELDS(node);

        spit_string(" ");
        spit_start_comp_stmt(SEQ);
        spit_newline();

        while (fields)
        {
          char *str;

          spit_string(indent_str);
          spit_decl(fields, NO_INDENT, 1);
          spit_string(";");
          spit_newline();
          fields = TREE_CHAIN(fields);
        }

        spit_end_comp_stmt(SEQ);
        spit_string(";");
        spit_newline();
        break;

      case ENUMERAL_TYPE:
        spit_string("enum ");

        if (TYPE_NAME(node))
          spit_string(IDENTIFIER_POINTER(TYPE_NAME(node)));
        spit_string("\n{");
        values = TYPE_VALUES(node);

        while (values)
        {
          spit_newline();
          spit_string( /* indent_str */ "    ");
          spit_string(IDENTIFIER_POINTER(TREE_PURPOSE(values)));

          if (TREE_VALUE(values))
          {
            spit_string(" = ");
            spit_cst(TREE_VALUE(values));
          }
          values = TREE_CHAIN(values);
          if (values)
            spit_string(", ");
        }
        spit_newline();
        spit_string("};");
        spit_newline();
        break;
      default:
        fprintf(stderr, "Wrong opcode in spit_all_structures\n");
        fputs(tree_code_name[(int) TREE_CODE(node)], stderr);
        exit(0);

      }
    }
    temp = TREE_CHAIN(temp);
  }
  fprintf(c_file, "\n</PRE>\n");
  fprintf(c_file, "</HTML>\n");
}

/** spits out a structure without any name. This is output in the       **/
/** c-dump file itself, not in the structures file.                     **/
LOCAL void
spit_tagless_struct(tree node)
{
  tree  fields;

  if (TREE_CODE(node) == RECORD_TYPE)
    spit_string("struct ");
  else
    spit_string("union ");
  fields = TYPE_FIELDS(node);
  spit_start_comp_stmt(SEQ);
  while (fields)
  {
    char *str;
    spit_string(indent_str);
    spit_decl(fields, NO_INDENT, 1);
    spit_string(";");
    spit_newline();
    fields = TREE_CHAIN(fields);
  }
  spit_end_comp_stmt(SEQ);
  spit_string(";");
}

/** spits out name of a .._decl node                            **/
/* modified by C Donawa 3rd Feb '93.  For structures with bit padding i.e.

struct{ int twobit:2; int       :1; int threebit:3; int onebit:1; } s3;

then the bit field_decl node for the second field will have no name.  I
   simply check to see if DECL_NAME is NULL before printing the name. */


void
spit_name(node)
  tree  node;
{
  if (DECL_NAME(node))
  { if (dump_cur_func && (TREE_CODE(node) == FUNCTION_DECL) &&
        (!(IS_PSEUDO_FUNCTION(node) || TREE_EXTERNAL(node))))
    {
      spit_html_link_string(c_file, node, IDENTIFIER_POINTER(DECL_NAME(node)),
			    c_base_name, ".html", "source");
    }
#if 0 /* a TYPE_DECL node doesn't always represent a  typedef: so give up */
    /* sometimes even a "float" type node has TREE_CODE TYPE_DECL: second
     * condition safeguards against this(I think the earthlings did this
     * confusing thing : Rakesh: Dec 18/96 */
    else if ((TREE_CODE(node) == TYPE_DECL) && (DECL_TYPE_NAME(node) != NULL))
    {
      spit_html_link_string(c_file, node, IDENTIFIER_POINTER(DECL_NAME(node)),
			    c_base_name, ".typedefs.html", NULL);
    }
#endif
    else
     spit_string(IDENTIFIER_POINTER(DECL_NAME(node)));
  } 
}


/** spits out a list of parameters during a function definition **/
LOCAL void
spit_params(params)
  tree  params;
{
  if (dump_func_prototype == 1)
  {
    while (params)
    {
     char *str;
 
     spit_decl(params, NO_INDENT, 1);
     params = TREE_CHAIN(params);
 
     if (params)
     {
       spit_string(", ");
 
       if (current_column >  64 )
       {
         spit_newline(); 
         spit_n_spaces(save_column); 
       }
      }
    }
  }
  else
  {
    while (params)
    {
     char *str;
 
     spit_newline();
     spit_decl(params, NO_INDENT, 1);
     params = TREE_CHAIN(params);
 
     if (params)
     {
       spit_string(", ");
 
      /* if (current_column >  64 ) Rakesh : Jan 27/97*/
       {
         /* spit_newline(); */
         /* spit_n_spaces(save_column);  Rakesh : Jan 27/97*/
       }
      }
    }
  }
}

LOCAL void
/** function prototypes                                         **/
spit_func_prototype(tree func_decl, int indent_flag)
{
#if 0
  if (TREE_LOCAL(func_decl))
  {
    if (dump_EARTH_comments)
      spit_string("/* local (basic) */ ");
    else if (EARTH_dump)
      spit_string("local /* basic */");
  }
  if (TREE_EXTERNAL(func_decl) && !TREE_LOCAL(func_decl))
    spit_string("extern ");

  if (!TREE_PUBLIC(func_decl) && TREE_STATIC(func_decl))
    spit_string("static ");
#endif

  if (indent_flag == INDENT)
    spit_string(indent_str);
  /**
  * We don't want qualifiers for function names?
  * spit_storage_class(func_decl);
  **/

  /* spit_func_header(func_decl); */

  /* Rakesh : Dec 10/96: flag to indicate we are dumping a func prototype */
  dump_func_prototype = 1;
  TYPE_ARG_LIST(TREE_TYPE(func_decl)) = DECL_ARGUMENTS(func_decl);
  spit_storage_class(func_decl);
  spit_decl(func_decl, NO_INDENT, 1);

  spit_string(";");
  spit_newline();
  dump_func_prototype = 0;
}

LOCAL void
spit_start_comp_stmt(End_type end_type)
{
  if (!EARTH_dump)
    end_type = SEQ;

  switch (end_type)
  {
  case SEQ:
    spit_indent_string("{");
    break;

  case PAR:
    spit_indent_string("{^");
    break;

  case STRICT_SEQ:
    spit_indent_string("{!");
    break;
  }
  inc_indent();
}

LOCAL void
spit_end_comp_stmt(End_type end_type)
{
  if (end_type == EMPTY)
    return;

  if (!EARTH_dump)
    end_type = SEQ;

  dec_indent();

  switch (end_type)
  {
  case SEQ:
    spit_indent_string("}");
    break;

  case PAR:
    spit_indent_string("^}");
    break;

  case STRICT_SEQ:
    spit_indent_string("!}");
    break;

  default:                      /* LJH */
    spit_indent_string("/* LJH */ }");
    break;
  }
  spit_newline();
}

void
dec_indent(void)
{
  int   id;

  curr_indent -= INDENT_STEP;

  if (curr_indent > INDENT_size)
    id = INDENT_size - 1;
  else
    id = curr_indent;

  indent_str[id] = '\0';
}

void
inc_indent(void)
{
  int   i;
  int   id;

  curr_indent += INDENT_STEP;

  if (curr_indent > INDENT_size)
    id = INDENT_size - 1;
  else
    id = curr_indent;

  for (i = 0; i < id; i++)
    indent_str[i] = ' ';

  indent_str[id] = '\0';
}

LOCAL
int
file_or_string(set)
  int   set;
{
  static file_or_string_flag = NO_STRING;

  if (set == STRING_SET)
    file_or_string_flag = STRING;
  else if (set == NO_STRING_SET)
    file_or_string_flag = NO_STRING;
  else
    return file_or_string_flag;
}


void
print_file_or_string(char *str)
{
  char *temp;

  if (file_or_string(VALUE) == STRING)
  {
    int   t1 = strlen(str) + 1;
    int   t2 = 0;
    /* VC *** */
    if (comment_string != NULL)
      t2 = strlen(comment_string) + 1;
    /****** Dont know why, realloc didn't work !! **/
    /****** Always returned NULL                  **/
    /*******
    temp = (char *) realloc (comment_string, t1+t2);
    if (temp == NULL)
        {
        fprintf(stderr, "ERROR!! No memory space available\n");
        exit(0);
        }
    comment_string = temp;
    ********/

    temp = (char *) Mmalloc(t1 + t2 + 5);
    if (temp == NULL)
    {
      fprintf(stderr, "ERROR!! No memory space available\n");
      exit(0);
    }
    /* sprintf(temp, "%s%s", comment_string, str); If comment string i s
       null, then "(null)" is printed, overwriting temp, as not enough memor
       y has been allocated */

    sprintf(temp, "%s%s", comment_string ? comment_string : "", str);

    /* printf ("  temp = %s\n", temp); strcpy(temp, comment_string);
       strcat(temp, str); */
    /*** VC ***/
    if (comment_string != NULL)
      Mfree(comment_string);
    comment_string = temp;
  }
  else
  {
    /* printf ("%s\n", str); */

    /* Here is the code which handles the folding of long lines.  The global
       variable `flag_fold_and_align' is a hack for passing an extra
       parameter to this routine.  If set, we determine whether the string to
       be printed will overflow the right margin, and if so, we go to a new
       line and indent to the saved column, which is presumably set to a
       sensible value.  For example, when a function call is dumped, the
       column saved is the one right after the opening argument list
       parenthesis.
    
    Note: be careful about using `spit_string()' and similar functions in
       this function (`print_file_or_string()'); they call this function, so
       you could introduce infinite recursion. */
    if (flag_fold_and_align)
    {
      int   len = strlen(str);

      /* Check for overflow past right margin. */
      if (len + current_column > right_margin)
      {
        fprintf(c_file, "\n");
        current_column = 0;

        /* If the string is very long, indentation would just throw the
           string beyond the right margin, so indent lightly in that case. */
        if (len + save_column < right_margin)
          spit_n_spaces(save_column);
        else
          spit_n_spaces(2);
      }
    }

    {
      int   column_increment = get_nb_columns_after_last_newline(str);

      /* If `str' contains a newline, `column_increment' will be less than
         the length of the string. */
      if (column_increment < strlen(str))
        current_column = 0;

      current_column += column_increment;
      fprintf(c_file, "%s", str);
    }
  }
  return;

}

static int
get_nb_columns_after_last_newline(char *str)
{
  int   len = strlen(str);
  int   i;

  for (i = len - 1; i >= 0 && str[i] != '\n'; i--);

  return len - i - 1;
}

static void
spit_n_spaces(int n)
{
  int   i;

  for (i = 0; i < n; i++)
    fprintf(c_file, " ");

  current_column += n;
}

void
spit_newline(void)
{
  print_file_or_string("\n");
}

void
spit_string(char *str)
{
  print_file_or_string(str);
}

/* Rakesh: Dec 11/96: Written to avoid advancing the columns for
 * strings dumped just for html formatting, and which don't form
 * part of the actual dumped code: this keeps the formatting intact
 * when the file is viewed via an html browser */
void
spit_html_string(char *str)
{
 int html_save_column, html_current_column;
 html_save_column = save_column;
 html_current_column = current_column;
 spit_string(str);
 save_column = html_save_column;
 current_column = html_current_column;
}

/* Rakesh: Dec 13/96: Dump the TREE_UID as a string to serve
 * as a unique html ref label for this node */
void
spit_html_uid_string(FILE *dump_file, tree node)
{
 FILE *t_file;
 int node_uid;
 char str[10]; /* node_uid shouldn't reach 10 digits!! */
 t_file = c_file;
 c_file = dump_file;
 node_uid = TREE_UID(node);
 sprintf(str, "%d", node_uid);
 spit_html_string(str);
 c_file = t_file;
}

/* Rakesh: Dec 13/96 : prints a _DECL node as an html target with
 * its TREE_UID as the ref label */
void
spit_html_target_string(FILE *dump_file, tree node, char *node_name)
{
 FILE *t_file;
 /* Useful when called from other modules and they have a different dump_file*/
 t_file = c_file;
 c_file = dump_file; 
 spit_html_string("<A NAME=\"");
 spit_html_uid_string(dump_file, node);
 spit_html_string("\">");
 if (node_name != NULL) /* For some targets we don't want to print any names */
  spit_string(node_name);
 spit_html_string("</a>");
 c_file = t_file;
}

/* Rakesh Dec 13/96: Dumps a _DECL node with html links. If file_ext
 * is non-null then it sets a cross-file html link : again TREE_UID
 * is used as the unique ref label*/
void spit_html_link_string(FILE *dump_file, tree node, char *node_name, 
		           char *base_file_name, char *file_ext, char *target)
{
 FILE *t_file;
 /* Useful when called from other modules and they have a different dump_file*/
 t_file = c_file;
 c_file = dump_file; 
 if (target != NULL)
 {
  spit_html_string("<A TARGET = \"");
  spit_html_string(target);
  spit_html_string("\"  HREF=\"");
 }
 else
  spit_html_string("<A HREF=\"");
 if (base_file_name != NULL) 
 {
  if (file_ext == NULL) 
  {
   spit_html_string(base_file_name);
   spit_html_string(".html");
  }
  else
  {
   spit_html_string(full_path_name);
   spit_html_string("/");
   spit_html_string(base_file_name);
   spit_html_string(file_ext);
  }
 }
 spit_html_string("#");
 spit_html_uid_string(dump_file, node);
 spit_html_string("\">");
 if (TREE_CODE(node) == FUNCTION_DECL)
 {
   if (file_ext != NULL)
    spit_html_string("<FONT color=blue>");
   else if (dump_func_prototype == 1)
    spit_html_string("<FONT color=green>");
   else
    spit_html_string("<FONT color=red>");
   spit_string(node_name);
   spit_html_string("</FONT></a>");
 }
 else
 {
   spit_string(node_name);
   spit_html_string("</a>");
 }
 c_file = t_file;
}

/* Rakesh Jan 16/96: Dumps a semi html link: till before the name of 
   the link: which is normally some C code. If file_ext
 * is non-null then it sets a cross-file html link : again TREE_UID
 * is used as the unique ref label*/
void spit_html_semi_link_string(FILE *dump_file, tree node,  
                 char *base_file_name, char *file_ext, char *target)
{
 FILE *t_file;
 /* Useful when called from other modules and they have a different dump_file*/
 t_file = c_file;
 c_file = dump_file; 
 if (target != NULL)
 {
  spit_html_string("<A TARGET = \"");
  spit_html_string(target);
  spit_html_string("\"  HREF=\"");
 }
 else
  spit_html_string("<A HREF=\"");
 if (file_ext != NULL)
 {
   spit_html_string(base_file_name);
   spit_html_string(file_ext);
 }
 spit_html_string("#");
 spit_html_uid_string(dump_file, node);
 spit_html_string("\">");
 c_file = t_file;
}

void spit_html_cgi_link_string(FILE *dump_file, tree node,  
                 char *base_file_name, char *file_ext, char *target)
{
 FILE *t_file;
 /* Useful when called from other modules and they have a different dump_file*/
 t_file = c_file;
 c_file = dump_file; 
 if (target != NULL)
 {
  spit_html_string("<A TARGET = \"");
  spit_html_string(target);
  spit_html_string("\"  HREF=\"");
 }
 else
  spit_html_string("<A HREF=\"");
 if (file_ext != NULL)
 {
   spit_html_string(base_file_name);
   spit_html_string(file_ext);
 }
 /*
 */
 spit_html_string("cgi-bin/ghiya/display.pl?");
 spit_html_uid_string(dump_file, node);
 spit_html_string("\">");
 if (TREE_CODE(node) == FUNCTION_DECL)
  spit_html_string("<FONT color=magenta>");
 else 
 {
  if (TREE_CODE(node) != EXPR_STMT)
  spit_html_string("<FONT color=blue>");
 }
 c_file = t_file;
}
void
spit_indent_string(char *str)
{
  char *temp = (char *) Mmalloc(strlen(indent_str) + strlen(str) + 3);
  sprintf(temp, "%s%s", indent_str, str);
  print_file_or_string(temp);
}

/** Recursive routine to spit out c code for a single statement **/
LOCAL void
spit_stmt(tree node, End_type end_type)
{
  extern char deadcode[];
  extern char notbasic[];
  /* extern char badnest[]; */

  /* We have to dump the correct type of closing brace for EARTH-C */
  if (node == NULL)
  {
    if (end_type != EMPTY)
    {
      if (dump_EARTH_comments)
      {
        spit_newline();
        spit_indent_string("/** End of ");

        switch (end_type)
        {
        case SEQ:
          spit_string("SEQ_LIST");
          break;
        case PAR:
          spit_string("PAR_LIST");
          break;
        case STRICT_SEQ:
          spit_string("STRICT_LIST");
          break;
        }
        spit_string(" **/\n");
      }
      spit_end_comp_stmt(end_type);
    }
    return;
  }
  dh_hook(DH_BEFORESTMT, node);

  switch (TREE_CODE(node))
  {
  case TREE_LIST:
    if (TREE_PURPOSE(node) != NULL_TREE
      || TREE_VALUE(node) != NULL_TREE
      && TREE_CODE(TREE_VALUE(node)) == TREE_LIST)
    {
      end_type = SEQ;           /* Set default value. */

      switch (TREE_PAR(node))
      {
      case REGULAR_C:
        spit_start_comp_stmt(SEQ);
        end_type = SEQ;
        break;

      case PARALLEL_C:
        spit_start_comp_stmt(PAR);
        end_type = PAR;
        break;

      case STRICT_SEQ_C:
        spit_start_comp_stmt(STRICT_SEQ);
        end_type = STRICT_SEQ;
        break;
      }
      spit_newline();

      if (TREE_COMMENT_STRING(node) != NULL)
      {
        spit_indent_string("/** ");
        spit_string(TREE_COMMENT_STRING(node));
        spit_string(" **/\n");
      }
      if (TREE_PURPOSE(node) != NULL_TREE)
      {
        spit_vars(STMT_VARS(TREE_PURPOSE(node)));
      } 
      else
      {
        if (TREE_COMMENT_STRING(node) == deadcode)
        {
          spit_indent_string(TREE_COMMENT_STRING(node));
        }
        if (TREE_COMMENT_STRING(node) == notbasic)
        {
          spit_indent_string(TREE_COMMENT_STRING(node));
        }
        /* if (TREE_COMMENT_STRING(node) == badnest); {
           spit_indent_string(TREE_COMMENT_STRING(node)); } */

      }
      if (dump_EARTH_comments)
      {
        switch (TREE_PAR(node))
        {
        case REGULAR_C:
          spit_indent_string("/** SEQ_LIST **/\n");
          end_type = SEQ;
          break;

        case PARALLEL_C:
          spit_indent_string("/** PAR_LIST **/\n");
          end_type = PAR;
          break;

        case STRICT_SEQ_C:
          spit_indent_string("/** STRICT_SEQ_LIST **/\n");
          end_type = STRICT_SEQ;
          break;
        }
      }
    }
    spit_stmt(TREE_VALUE(node), EMPTY);
    spit_stmt(TREE_CHAIN(node), end_type);
    break;

  case EXPR_STMT:
    spit_string(indent_str);
    spit_html_cgi_link_string(c_file, node, NULL, NULL, NULL);
    spit_expr(STMT_BODY(node), 0, node, 0);
    spit_html_string("</a>");
    spit_string(";");
    spit_newline();
    break;
  case IF_STMT:
    if (!is_fake_if(node))
    {
      spit_html_cgi_link_string(c_file, node, NULL, NULL, NULL);
      spit_indent_string("if (");
      dh_hook(DH_BEFORECOND, STMT_COND(node));
      spit_expr(STMT_COND(node), 0, node, SSA_RHS);
      dh_hook(DH_AFTERCOND, STMT_COND(node));
      spit_string(")");
      spit_html_string("</FONT></a>");
    }
    if (STMT_THEN(node) != NULL_TREE)
    {
      spit_newline();
      spit_stmt(STMT_THEN(node), EMPTY);
    }
    else
    {
      spit_string(";");
      spit_newline();
    }
    if (STMT_ELSE(node) != NULL_TREE)
    {
      spit_html_string("<FONT color=blue>");
      spit_indent_string("else");
      spit_html_string("</FONT>");
      spit_newline();
      spit_stmt(STMT_ELSE(node), EMPTY);
    }
    break;
  case WHILE_STMT:
    spit_html_cgi_link_string(c_file, node, NULL, NULL, NULL);
    spit_indent_string("while ( ");
    dh_hook(DH_BEFORECOND, STMT_WHILE_COND(node));
    spit_expr(STMT_WHILE_COND(node), 0, node, SSA_RHS);
    dh_hook(DH_AFTERCOND, STMT_WHILE_COND(node));
    spit_string(" )");
    spit_html_string("</FONT></a>");

    if (STMT_BODY(node) == NULL)
    {
      spit_string(";");
      spit_newline();
    }
    else
    {
      spit_newline();
      spit_stmt(STMT_BODY(node), EMPTY);
    }
    break;
  case DO_STMT:
    spit_html_cgi_link_string(c_file, node, NULL, NULL, NULL);
    spit_indent_string("do ");
    spit_html_string("</FONT></a>");

    if (STMT_BODY(node) != NULL)
    {
      spit_newline();
      spit_stmt(STMT_BODY(node), EMPTY);
    }
    spit_html_string("<FONT color=blue>");
    spit_indent_string("while ( ");
    dh_hook(DH_BEFORECOND, STMT_DO_COND(node));
    spit_expr(STMT_DO_COND(node), 0, node, SSA_RHS);
    dh_hook(DH_AFTERCOND, STMT_DO_COND(node));
    spit_string(" );");
    spit_html_string("</FONT>");
    spit_newline();
    break;
  case FOR_STMT:
    spit_html_cgi_link_string(c_file, node, NULL, NULL, NULL);
    if (IS_FORALL(node))
    {
      if (dump_EARTH_comments)
        spit_indent_string("for/*all */ ( ");
      else if (EARTH_dump)
        spit_indent_string("forall ( ");
      else
        spit_indent_string("for ( ");
    }
    else
      spit_indent_string("for ( ");

    dh_hook(DH_BEFOREINIT, STMT_START(node));
    spit_expr(STMT_START(node), 0, STMT_START(node), 0);
    dh_hook(DH_AFTERINIT, STMT_START(node));
    spit_string(" ; ");
    dh_hook(DH_BEFORECOND, STMT_END(node));
    spit_expr(STMT_END(node), 0, STMT_END(node), SSA_RHS);
    dh_hook(DH_AFTERCOND, STMT_END(node));
    spit_string(" ; ");
    dh_hook(DH_BEFOREITER, STMT_ITER(node));
    spit_expr(STMT_ITER(node), 0, STMT_ITER(node), 0);
    dh_hook(DH_AFTERITER, STMT_ITER(node));
    spit_string(" )");
    spit_html_string("</FONT></a>");

    if (STMT_BODY(node) == NULL)
    {
      spit_string(";");
      spit_newline();
    }
    else
    {
      spit_newline();
      spit_stmt(STMT_BODY(node), EMPTY);
    }
    break;
  case CONTINUE_STMT:
    spit_indent_string("continue; ");
    spit_newline();
    break;
  case BREAK_STMT:
    spit_indent_string("break; ");
    spit_newline();
    break;
  case RETURN_STMT:
    spit_indent_string("return");
    if (STMT_BODY(node))
    {
      spit_string(" (");
      spit_expr(STMT_BODY(node), 0, node, SSA_RHS);
      spit_string(")");
    }
    spit_string(";");
    spit_newline();
    break;
  case GOTO_STMT:
    spit_indent_string("goto ");
    spit_expr(STMT_BODY(node), 0, NULL, 0);
    spit_string(";");
    spit_newline();
    break;

  case LABEL_STMT:
    spit_newline();
    spit_expr(STMT_BODY(node), 0, NULL, 0);
    spit_string(":;");
    spit_newline();
    break;

  case SWITCH_STMT:
    spit_html_cgi_link_string(c_file, node, NULL, NULL, NULL);
    spit_indent_string("switch ");
    spit_string("(");
    dh_hook(DH_BEFORECOND, STMT_SWITCH_EXPR(node));
    spit_expr(STMT_SWITCH_EXPR(node), 0, node, SSA_RHS);
    dh_hook(DH_AFTERCOND, STMT_SWITCH_EXPR(node));
    spit_string(")");
    spit_html_string("</FONT></a>");
    spit_newline();
    spit_stmt(STMT_SWITCH_STMT(node), EMPTY);
    break;

  case CASE_STMT:
    {
      tree  temp = STMT_CASE_EXPR(node);
      if (TREE_CODE(temp) == TREE_LIST)
      {
        /** After SIMPLIFY routine, the case**/
        /** expressions are connected   **/
        /** by treelists.                       **/
        /** There could be more than 1 expr **/
        /** in a case statement         **/
        while (temp)
        {
          spit_html_cgi_link_string(c_file, temp, NULL, NULL, NULL);
          spit_indent_string("case ");
          spit_expr(TREE_VALUE(temp), 0, NULL, 0);
          spit_string(" : ");
          spit_html_string("</FONT></a>");
          spit_newline();
          temp = TREE_CHAIN(temp);
        }
      }
      else
      {
        spit_html_cgi_link_string(c_file, temp, NULL, NULL, NULL);
        spit_indent_string("case ");
        spit_expr(temp, 0, NULL, 0);
        spit_string(" : ");
        spit_html_string("</FONT></a>");
        spit_newline();
      }
      inc_indent();
      spit_stmt(STMT_CASE_STMT(node), EMPTY);
      dec_indent();
      break;
    }

  case DEFAULT_STMT:
    spit_html_cgi_link_string(c_file, node, NULL, NULL, NULL);
    spit_indent_string("default : ");
    spit_html_string("</FONT></a>");
    spit_newline();
    inc_indent();
    spit_stmt(STMT_DEFAULT_STMT(node), EMPTY);
    dec_indent();
    break;

  default:
    fprintf(stderr, "Wrong opcode in spit_stmt\n");
    fputs(tree_code_name[(int) TREE_CODE(node)], stderr);
    exit(0);
  }

  dh_hook(DH_AFTERSTMT, node);
}

/** This routine was put in in order to take care of                    **/
/** pointer arithmetic. For example, if a constant is used in pointer   **/
/** arithmetic, it should be divided by the size of the pointer element **/
/** The  mod flag have the same meaning as in the routine               **/
/** spit_expr. There is a static variable -- addr_type -- which keeps   **/
/** track of the size of the objects we are dealing with during the     **/
/** expression evaluation. This is used to put out the correct constant **/
/** values                                                              **/

/** c-ast-c --- Since the tree built was modified to take care of       **/
/** pointer arithmetic, this routine is no longer necessary-- so        **/
/** it just calls spit_operands, and does nothing else                  **/
LOCAL void
spit_check_operands(tree node, char *str, int mod_flag, int no_of_ops,
  tree stmt, int side)
{
  static tree addr_type = NOT_SET;

  if (addr_type == NOT_SET)
  {
    if (TREE_CODE(node) == INTEGER_CST)
    {
      spit_cst(node);
      return;
    }
    if (TREE_CODE(TREE_TYPE(node)) == POINTER_TYPE)
      addr_type = TYPE_SIZE(TREE_TYPE(TREE_TYPE(node)));
    if (no_of_ops == 1)
      spit_operand(node, str, mod_flag, stmt, side);
    else
      spit_operands(node, str, mod_flag, stmt, side);
    addr_type = NOT_SET;
  }
  else
  {
    if (TREE_CODE(node) == INTEGER_CST)
    {
      spit_cst(node);
      /** c-ast-c **/
      /**
      spit_string("/");
      spit_cst(addr_type);
      **/
    }
    else
    {
      if (no_of_ops == 1)
        spit_operand(node, str, mod_flag, stmt, side);
      else
        spit_operands(node, str, mod_flag, stmt, side);
    }
  }
}



/** spits out operands for a binary operation                   **/
LOCAL void
spit_operands(tree node, char *str, int mod_flag, tree stmt, int side)
{
  if (mod_flag)
    spit_string("(");

  spit_expr(TREE_OPERAND(node, 0), 1, stmt, side);
  spit_string(" ");
  spit_string(str);
  spit_string(" ");
  spit_expr(TREE_OPERAND(node, 1), 1, stmt, side);

  if (mod_flag)
    spit_string(")");
}

/** spits out code for an unary operation               **/
LOCAL void
spit_operand(tree node, char *str, int mod_flag, tree stmt, int side)
{
  if (mod_flag)
    spit_string("(");

  spit_string(str);
  /* spit_string(" "); */
  spit_expr(TREE_OPERAND(node, 0), 1, stmt, side);

  if (mod_flag)
    spit_string(")");
}


/** recursive routine to spit out c code for expressions        **/
/** 'mod_flag' is to determine whether to put brackets          **/
/** around modify_expr.                                         **/
LOCAL void
spit_expr(tree node, int mod_flag, tree stmt, int side)
{
  char *str;
  enum tree_code code;

  if (node == NULL_TREE)
    return;

  dh_hook(DH_BEFOREEXPR, node);

  code = TREE_CODE(node);
  switch (code)
  {
  case CONSTRUCTOR:
    {
      tree  temp;

      spit_string("{");
      temp = TREE_OPERAND(node, 1);
      while (temp)
      {
        spit_expr(TREE_VALUE(temp), 0, NULL, 0);
        temp = TREE_CHAIN(temp);
        if (temp)
          spit_string(",");
        spit_newline();
      }
      spit_string("}");
      break;
    }

  case MODIFY_EXPR:

    if (mod_flag)
      spit_string("(");
    spit_expr(TREE_OPERAND(node, 0), 1, stmt, SSA_LHS);
    spit_string(" = ");
    spit_expr(TREE_OPERAND(node, 1), 1, stmt, SSA_RHS);
    if (mod_flag)
      spit_string(")");
    break;

  case PLUS_EXPR:
    spit_check_operands(node, "+", 1, 2, stmt, side);
    break;
  case MINUS_EXPR:
    spit_check_operands(node, "-", 1, 2, stmt, side);
    break;
  case MULT_EXPR:
    spit_check_operands(node, "*", 1, 2, stmt, side);
    break;
  case TRUNC_DIV_EXPR:
    spit_check_operands(node, "/", 1, 2, stmt, side);
    break;
  case CEIL_DIV_EXPR:
    spit_check_operands(node, "/", 1, 2, stmt, side);
    break;
  case FLOOR_DIV_EXPR:
    spit_check_operands(node, "/", 1, 2, stmt, side);
    break;
  case ROUND_DIV_EXPR:
    spit_check_operands(node, "/", 1, 2, stmt, side);
    break;
  case TRUNC_MOD_EXPR:
    spit_check_operands(node, "%", 1, 2, stmt, side);
    break;
  case CEIL_MOD_EXPR:
    spit_check_operands(node, "%", 1, 2, stmt, side);
    break;
  case FLOOR_MOD_EXPR:
    spit_check_operands(node, "%", 1, 2, stmt, side);
    break;
  case ROUND_MOD_EXPR:
    spit_check_operands(node, "%", 1, 2, stmt, side);
    break;
  case FIX_TRUNC_EXPR:
    spit_check_operands(node, "trunc", 1, 1, stmt, side);
    break;
  case FIX_CEIL_EXPR:
    spit_check_operands(node, "ceil", 1, 1, stmt, side);
    break;
  case FIX_FLOOR_EXPR:
    spit_check_operands(node, "floor", 1, 1, stmt, side);
    break;
  case FIX_ROUND_EXPR:
    spit_check_operands(node, "round", 1, 1, stmt, side);
    break;
  case FLOAT_EXPR:
    spit_check_operands(node, "(float)", 1, 1, stmt, side);
    break;
  case NEGATE_EXPR:
    spit_check_operands(node, "-", 1, 1, stmt, side);
    break;
  case EXPON_EXPR:
    spit_check_operands(node, "exp", 1, 2, stmt, side);
    break;
  case MIN_EXPR:
    spit_check_operands(node, "min", 1, 2, stmt, side);
    break;
  case MAX_EXPR:
    spit_check_operands(node, "max", 1, 2, stmt, side);
    break;
  case RDIV_EXPR:
    spit_check_operands(node, "/", 1, 2, stmt, side);
    break;
  case EXACT_DIV_EXPR:
    spit_check_operands(node, "/", 1, 2, stmt, side);
    break;
  case COMPOUND_EXPR:
    /**********88
        spit_string("({");
        spit_expr(TREE_OPERAND(node,0),1);
        spit_string(";");
        spit_expr(TREE_OPERAND(node,1),1);
        spit_string(";");
        spit_string("})");
        **************/
    spit_expr(TREE_OPERAND(node, 0), 1, NULL, 0);
    spit_string(",");
    spit_expr(TREE_OPERAND(node, 1), 1, NULL, 0);

    break;
  case ABS_EXPR:
    spit_check_operands(node, "abs", 1, 1, stmt, side);
    break;
  case LSHIFT_EXPR:
    spit_check_operands(node, "<<", 1, 2, stmt, side);
    break;
  case RSHIFT_EXPR:
    spit_check_operands(node, ">>", 1, 2, stmt, side);
    break;
  case BIT_IOR_EXPR:
    spit_check_operands(node, "|", 1, 2, stmt, side);
    break;
  case BIT_XOR_EXPR:
    spit_check_operands(node, "^", 1, 2, stmt, side);
    break;
  case BIT_AND_EXPR:
    spit_check_operands(node, "&", 1, 2, stmt, side);
    break;
  case TRUTH_ANDIF_EXPR:
  case TRUTH_AND_EXPR:
    spit_check_operands(node, "&&", 1, 2, stmt, side);
    break;
  case TRUTH_ORIF_EXPR:
  case TRUTH_OR_EXPR:
    spit_check_operands(node, "||", 1, 2, stmt, side);
    break;
  case LT_EXPR:
    spit_check_operands(node, "<", 1, 2, stmt, side);
    break;
  case LE_EXPR:
    spit_check_operands(node, "<=", 1, 2, stmt, side);
    break;
  case GT_EXPR:
    spit_check_operands(node, ">", 1, 2, stmt, side);
    break;
  case GE_EXPR:
    spit_check_operands(node, ">=", 1, 2, stmt, side);
    break;
  case EQ_EXPR:
    spit_check_operands(node, "==", 1, 2, stmt, side);
    break;
  case NE_EXPR:
    spit_check_operands(node, "!=", 1, 2, stmt, side);
    break;

  case ARRAY_REF:
    spit_expr(TREE_OPERAND(node, 0), 1, stmt, side);
    spit_string("[");
    spit_expr(TREE_OPERAND(node, 1), 1, stmt, SSA_RHS);
    spit_string("]");
    break;

  case COMPONENT_REF:
    spit_expr(TREE_OPERAND(node, 0), 1, stmt, side);
    spit_string(".");
    spit_expr(TREE_OPERAND(node, 1), 1, NULL, 0);
    spit_SSA_Number(node, stmt, side);
    break;

  case BIT_NOT_EXPR:
    spit_check_operands(node, "~", 1, 1, stmt, side);
    break;

  case TRUTH_NOT_EXPR:
    spit_check_operands(node, "!", 1, 1, stmt, side);
    break;

  case CONVERT_EXPR:
  case NOP_EXPR:
    spit_type_cast(node);
    spit_string(" ");
    spit_expr(TREE_OPERAND(node, 0), 1, stmt, side);
    break;

  case SAVE_EXPR:
    spit_expr(TREE_OPERAND(node, 0), 1, NULL, side);
    break;

  case PREDECREMENT_EXPR:
    spit_check_operands(node, "--", 1, 1, stmt, side);
    break;

  case PREINCREMENT_EXPR:
    spit_check_operands(node, "++", 1, 1, stmt, side);
    break;

  case POSTDECREMENT_EXPR:
    spit_string("(");
    spit_expr(TREE_OPERAND(node, 0), 1, stmt, side);
    spit_string("--");
    spit_string(")");
    break;

  case POSTINCREMENT_EXPR:
    spit_string("(");
    spit_expr(TREE_OPERAND(node, 0), 1, stmt, side);
    spit_string("++");
    spit_string(")");
    break;

  case ADDR_EXPR:
    /** Rakesh: Nov 20/96: Hack to handle cases like int t[100], (*ptr)[100],
       *rt; S1: ptr = &t; S2: rt = t; Case S1 and S2 are different: in
       c-dump, we _do_ need to dump the "&" operator for S1: in C t and &t[0]
       are equivalent but not t and &t: well &t represents the same location
       but the type of this expression is not pointer to integer but pointer
       to an array of integer. But the ASTs for stmts S1 and S2 are
       identical: in both cases we have an ADDR_EXPR appended before the
       VAR_DECL t. However the type of the ADDR_EXPR itself differs: for S1
       it is pointer type to array type of integer type, while for S2 it is
       pointer type to integer type. So we use the type of the ADDR_EXPR to
       distinguish between the two cases and dump the "&" operator for the S1
       case. Also need to check that we are not dumping "&" operator
       in front of an array-ref: function parent_is_array_ref is called
       for that **/
   if ((TREE_CODE(TREE_OPERAND(node, 0)) == STRING_CST))
      spit_expr(TREE_OPERAND(node, 0), 0, stmt, side);
   else if (TREE_CODE(TREE_TYPE(TREE_OPERAND(node, 0))) == ARRAY_TYPE)
   {
      if ((!parent_is_array_ref(node, stmt, side)) &&
         (height_difference(TREE_TYPE(node), TREE_TYPE(TREE_OPERAND(node, 0)))))
       spit_check_operands(node, "&", 1, 1, stmt, side);
      else
       spit_expr(TREE_OPERAND(node, 0), 0, stmt, side);
    }
   else
      spit_check_operands(node, "&", 1, 1, stmt, side);
    break;

  case COND_EXPR:
    spit_string("(");
    spit_expr(TREE_OPERAND(node, 0), 1, stmt, SSA_RHS);
    spit_string("?");
    spit_expr(TREE_OPERAND(node, 1), 1, stmt, side);
    spit_string(":");
    spit_expr(TREE_OPERAND(node, 2), 1, stmt, side);
    spit_string(")");
    break;

  case CALL_EXPR:
    if (TREE_CODE(TREE_OPERAND(node, 0)) == ADDR_EXPR)
      spit_expr(TREE_OPERAND(TREE_OPERAND(node, 0), 0), 1, NULL, SSA_RHS);
    else
      /** this happens when we have a variable        **/
      /** pointing to a function                                      **/
      spit_expr(TREE_OPERAND(node, 0), 1, stmt, SSA_RHS);

    flag_fold_and_align = 0;
#if 1 /*  Rakesh : Jan 27/97 : use the uid of the stmt though*/
       /* only when doing c-dump for a simple call */
    if (dump_cur_func && (TREE_CODE(STMT_BODY(stmt)) == CALL_EXPR)) 
     spit_html_cgi_link_string(c_file, stmt, NULL, NULL, NULL);
#endif
    spit_call_expr_num(node);
    spit_string("(");
    save_column = current_column;
    {
      tree  params = TREE_OPERAND(node, 1);
      while (params)
      {
        spit_arg(params, stmt);
        params = TREE_CHAIN(params);
        if (params)
          spit_string(", ");
      }
    }
#if 1
    if (dump_cur_func && (TREE_CODE(STMT_BODY(stmt)) == CALL_EXPR)) 
     spit_html_string("</a>");
#endif
    spit_string(")");

    if (dump_EARTH_comments)
      spit_string(" /* ");

    if (flag_earth_syntax && (EARTH_dump || dump_EARTH_comments))
    {
      if (CALL_SITE(node) != NULL_TREE)
      {
        flag_fold_and_align = 1;
        spit_string(" @ ");
        spit_expr(CALL_SITE(node), 1, stmt, SSA_RHS);
        flag_fold_and_align = 0;
      }
    }
    if (dump_EARTH_comments)
      spit_string(" */");
    break;

  case TREE_LIST:
    spit_expr(TREE_VALUE(node), 1, stmt, side);
    break;

  case VAR_DECL:
  case CONST_DECL:
  case PARM_DECL:
  case FIELD_DECL:
  case RESULT_DECL:
  case FUNCTION_DECL:
  case LABEL_DECL:
    spit_name(node);
    spit_SSA_Number(node, stmt, side);
    break;

  case INTEGER_CST:
    spit_check_operands(node, "", 1, 1, NULL, 0);
    break;
  case REAL_CST:
  case COMPLEX_CST:
  case STRING_CST:
    spit_cst(node);
    break;

  case INDIRECT_REF:
    if (IS_ARRAY_REF(node))
    {
      spit_expr(TREE_OPERAND(node, 1), 1, stmt, side);
      spit_string("[");
      spit_expr(TREE_OPERAND(node, 2), 1, stmt, SSA_RHS);
      spit_string("]");
    }
    else
      spit_check_operands(node, "*", 1, 1, stmt, side);
    break;
  default:
    fprintf(stderr, "check_expr_nodes, Error in spit_expr, \n");
    fputs(tree_code_name[(int) TREE_CODE(node)], stderr);
    exit(0);
  }

  dh_hook(DH_AFTEREXPR, node);

}

/** spits out code for a function argument.                     **/
/** Entire arrays are passed by reference in C, and so the      **/
/** AST has an ADDR_EXPR if the entire array is passed,         **/
/** therefore proper care should be taken while outputting C    **/
/** code. Similarly for string constants passed as arguments    **/
LOCAL void
spit_arg(tree arg, tree stmt)
{
  tree  node;

  node = TREE_VALUE(arg);
  while (TREE_CODE(node) == NOP_EXPR)
    node = TREE_OPERAND(node, 0);
  /*****
  if (TREE_CODE(node) == ADDR_EXPR)
          {
          if (((TREE_CODE(TREE_OPERAND(node,0)) == VAR_DECL)
           ||  (TREE_CODE(TREE_OPERAND(node,0)) == PARM_DECL))
           && (TREE_CODE(TREE_TYPE(TREE_OPERAND(node,0))) == ARRAY_TYPE))
           spit_expr(TREE_OPERAND(node,0),1,1);
          if (TREE_CODE(TREE_OPERAND(node,0)) == STRING_CST)
           spit_expr(TREE_OPERAND(node,0),1,1);
           }
  else
          spit_expr(node);
          ****/

  spit_expr(node, 1, stmt, SSA_RHS);
}

/** returns the integer constant the node represents            **/
EXPORT
int
get_int_cst(node)
  tree  node;
{
  char  str[20];

  if (TREE_CODE(node) != INTEGER_CST)
  {
    fprintf(stderr, "array index not a constant\n");
    exit(0);
  }
  if (TREE_INT_CST_HIGH(node) == 0)
    sprintf(str, "%1u", TREE_INT_CST_LOW(node));
  else if (TREE_INT_CST_HIGH(node) == -1 && TREE_INT_CST_LOW(node) != 0)
    sprintf(str, "%1u", -TREE_INT_CST_LOW(node));
  else
    sprintf(str, "0x%x%08x", TREE_INT_CST_HIGH(node), TREE_INT_CST_LOW(node));
  return (atoi(str));
}

/** spits out constants                                         **/
void
spit_cst(node)
  tree  node;
{
  switch (TREE_CODE(node))
  {
  case INTEGER_CST:
    {
      char  str[20];
      if (TREE_INT_CST_HIGH(node) == 0)
        sprintf(str, "%1u", TREE_INT_CST_LOW(node));
      else if (TREE_INT_CST_HIGH(node) == -1 && TREE_INT_CST_LOW(node) != 0)
        sprintf(str, "-%1u", -TREE_INT_CST_LOW(node));
      else
        sprintf(str, "0x%x%08x", TREE_INT_CST_HIGH(node), TREE_INT_CST_LOW(node));
      print_file_or_string(str);
    }
    break;
  case REAL_CST:
#ifndef REAL_IS_NOT_DOUBLE
    {
      /** Bhama **/
      /* The largest number -- a double is 2^64 -- which will  take up 20
         characters */
      char *str = (char *) Mmalloc(25);
      sprintf(str, "%.14e", TREE_REAL_CST(node));
      print_file_or_string(str);
    }
#else
    /**********8
    Bhama-- c-ast-c-- REAL_IS_NOT_DOUBLE is not set, hence to make
    the comment_printing job easier, this has been commented out...
                    {
                      int i;
                      char *p = (char *) &TREE_REAL_CST (node);
                      fprintf (c_file, "0x");
                      for (i = 0; i < sizeof TREE_REAL_CST (node); i++)
                      fprintf (c_file, "%02x", *p++);
                      }
                      ***************/
#endif                          /* REAL_IS_NOT_DOUBLE */
    break;
  case STRING_CST:
    {
      char *ch;
      char *str;
      int   i = 0;
      int   len = strlen(TREE_STRING_POINTER(node));

      /* I'm adding 10 for safetys sake.  When converting control characters,
         len is 1, but the control character is replaced by a '\' character
         and the char rep itself (like \n -> '\' and 'n'.  Chris Donawa, 19th
         Feb 1994 */

      str = (char *) Mmalloc(2 * len + 10);

      /**
            if (file_or_string(VALUE) == STRING)
            {
              print_file_or_string(TREE_STRING_POINTER(node));
              return;
            }
                **/
      ch = TREE_STRING_POINTER(node);
      sprintf(str, "\"");
      while (ch[i] != '\0')
      {
        switch (ch[i])
        {
        case '\n':
          sprintf(str, "%s\\n", str);
          break;
        case '\t':
          sprintf(str, "%s\\t", str);
          break;
        case '\b':
          sprintf(str, "%s\\b", str);
          break;
        case '"':
          sprintf(str, "%s\\\"", str);
          break;
        case '\\':
          sprintf(str, "%s\\\\", str);
          break;
        case '/':
          if (file_or_string(VALUE) == NO_STRING)
            sprintf(str, "%s%c", str, ch[i]);
          else if (ch[i + 1] != '*')
            sprintf(str, "%s%c", str, ch[i]);
          break;
        case '*':
          if (file_or_string(VALUE) == NO_STRING)
            sprintf(str, "%s%c", str, ch[i]);
          else if (ch[i + 1] != '/')
            sprintf(str, "%s%c", str, ch[i]);
          break;
        default:
          sprintf(str, "%s%c", str, ch[i]);
        }
        i++;
      }
      sprintf(str, "%s\"", str);
      print_file_or_string(str);
      str = (char *) Mfree(str);
      break;
    }
  default:
    fprintf(stderr, "check_expr_nodes, Error in spit_cst, \n");
    fputs(tree_code_name[(int) TREE_CODE(node)], stderr);
    exit(0);

  }
}

EXPORT void
get_comment_string(tree node)
{
  file_or_string(STRING_SET);
  spit_expr(node, 0, NULL, 0);
  file_or_string(NO_STRING_SET);
}

/* Determine whether `node' is a basic term, i.e., an expression which is not
   itself composed of an operator and subexpression. */
static int
is_basic_term(tree node)
{
  if (node == NULL_TREE)
    return 0;

  switch (TREE_CODE(node))
  {
  case INTEGER_CST:
  case REAL_CST:
  case STRING_CST:
  case VAR_DECL:
  case PARM_DECL:
    return 1;
  default:
    return 0;
  }
}

static int
basic_type(tree decl)
{
  if ((TREE_TYPE(decl) == integer_type_node) ||
    (TREE_TYPE(decl) == short_integer_type_node) ||
    (TREE_TYPE(decl) == long_integer_type_node) ||
    (TREE_TYPE(decl) == long_long_integer_type_node) ||
    (TREE_TYPE(decl) == short_unsigned_type_node) ||
    (TREE_TYPE(decl) == unsigned_type_node) ||
    (TREE_TYPE(decl) == long_unsigned_type_node) ||
    (TREE_TYPE(decl) == long_long_unsigned_type_node) ||

    (TREE_CODE(TREE_TYPE(decl)) == ENUMERAL_TYPE))
    return 1;
  else
    return 0;
}

/* -------------------------------------------------------------------- */
/* Stuff added to facilitate hooking in routines that add to the c-dump */

/* Insert a function into some hook-lists */
int
dh_insert(dh_func dhf, int code)
{
  dumphook *dh;
  int   wl;

  wl = 0;
  while (code)
  {

    /* Find list-index for each 1-bit of the code */
    while (!(code % 2) && code != 0)
    {
      code >>= 1;
      wl++;
    }

    if (!code)
      break;

    /* Insert dhf into wl dumphook-list */
    if (wl >= DH_MAX || wl < 0)
    {
      simple_warning(NULL, "dh_insert: Bad value for dumphook list");
      return (-1);
    }
    dh = (dumphook *) Mmalloc(sizeof(dumphook));
    if (dh == NULL)
      return (-1);

    dh->dhf = dhf;
    dh->next = dh_lists[wl];
    dh_lists[wl] = dh;

    /* Skip over 1-bit for this list and go on to next */
    code >>= 1;
    wl++;
  }

  return (0);
}


/* Remove a function from hook-list(s) */
int
dh_remove(dh_func dhf, int code)
{
  dumphook *dh, *lastdh;
  int   wl;

  wl = 0;
  while (code)
  {

    while (!(code % 2) && code != 0)
    {
      code >>= 1;
      wl++;
    }

    if (!code)
      break;

    /* Remove dhf from wl dumphook-list */
    if (wl >= DH_MAX || wl < 0)
    {
      simple_warning(NULL, "dh_remove: Bad value for dumphook list");
      return (-1);
    }
    dh = dh_lists[wl];
    lastdh = NULL;
    while (dh != NULL)
    {
      if (dh->dhf == dhf)
        break;
      lastdh = dh;
      dh = dh->next;
    }

    if (dh != NULL)
    {
      if (lastdh == NULL)
        dh_lists[wl] = dh->next;
      else
        lastdh->next = dh->next;

      dh = (dumphook *) Mfree(dh);
    }
    code >>= 1;
    wl++;
  }

  return (0);
}


/* Calls each of the functions sequentially in the given list. */
static int
dh_hook(int wl, tree node)
{
  dumphook *dh, *dhnext;

  if (wl >= DH_MAX || wl < 0)
  {
    simple_warning(node, "dh_insert: Bad value for dumphook list");
    return (-1);
  }
  dh = dh_lists[wl];
  while (dh != NULL)
  {
    dhnext = dh->next;
    if (dh->dhf != NULL)
    {
      (*(dh->dhf)) (node, curr_indent);
    }
    dh = dhnext;
  }

  return (0);
}

/***

Function `is_fake_if()':

 Check for the case

    if( 1 != 0 )
    {!
    !}

or
    if( 1 != 0 )
    {^
    ^}

and return a C boolean, i.e. an `int'.

==> Should be called only for a TREE_LIST node.

***/

int
is_fake_if(tree if_stmt)
{
  tree  stmt_cond, ne_expr, cond_left, cond_right;

  /* check if a valid if */
  if (if_stmt == NULL_TREE
    || TREE_CODE(if_stmt) != IF_STMT)
    return 0;

  /* LJH - check that the filename is "__FAKE_IF" */
  return ((STMT_SOURCE_FILE(if_stmt) != NULL) &&
    (strcmp(STMT_SOURCE_FILE(if_stmt), "__FAKE_IF") == 0)
    );
}

void
spit_record_type(tree type)
{
  if (TREE_CODE(type) == RECORD_TYPE)
    spit_string("struct ");

  else if (TREE_CODE(type) == UNION_TYPE)
    spit_string("union ");

  else
    spit_string("/* ?RECORD_TYPE */ ");

  if (TYPE_NAME(type) && TREE_CODE(TYPE_NAME(type)) == IDENTIFIER_NODE)
    spit_string(IDENTIFIER_POINTER(TYPE_NAME(type)));

  return;
}

static void
spit_enum_type(tree type)
{
  spit_string("enum ");

  if (TYPE_NAME(type) && TREE_CODE(TYPE_NAME(type)) == IDENTIFIER_NODE)
    spit_string(IDENTIFIER_POINTER(TYPE_NAME(type)));

  return;
}

static tree
get_base_type(tree type)
{
  tree  current = type;

  while (current)
  {
    if (TREE_CODE(current) == ARRAY_TYPE
      || TREE_CODE(current) == POINTER_TYPE
      || TREE_CODE(current) == FUNCTION_TYPE)
      current = TREE_TYPE(current);

    else
      break;
  }

  return current;
}

static void
spit_base_type(tree type)
{
  tree  qualified_type = type;

#if 0
  spit_string("/* spit_base_type() */ ");
#endif

  if (type == NULL_TREE)
    return;

  /* Warning: gcc does not produce any type tree such that
  
  TREE_CODE(type) == TYPE_DECL;
  
  only modules which are run later can produce this.  Do not produce such
     trees unless it is really necessary, as the type checker cannot handle
     type trees constructed in this way.  So far, only EARTH thread
     generation produces such trees. */
  if (TREE_CODE(type) == TYPE_DECL)
  {
    spit_name(type);
    spit_string(" ");

    if (!flag_earth_syntax)
      spit_string("/* warning: base type should not be a TYPE_DECL */ ");
    return;
  }

  type = TYPE_MAIN_VARIANT(type);       /* PO:1996-05-29 */

  switch (TREE_CODE(type))
  {
  case VOID_TYPE:
    spit_type_qual(qualified_type);

    spit_string("void");
    break;

  case CHAR_TYPE:
    spit_type_qual(qualified_type);

    if (type == signed_char_type_node)
      spit_string("signed ");

    else if (TREE_UNSIGNED(type))
      spit_string("unsigned ");

    spit_string("char");
    break;

  case INTEGER_TYPE:
    spit_type_qual(qualified_type);

    if (TREE_UNSIGNED(type))
      spit_string("unsigned ");

    if (type == short_integer_type_node
      || type == short_unsigned_type_node
      || TYPE_PRECISION(type) == SHORT_TYPE_SIZE)
      spit_string("short");

    else if (type == long_long_integer_type_node
        || type == long_long_unsigned_type_node
        || (TYPE_PRECISION(type) == LONG_LONG_TYPE_SIZE
        && (LONG_LONG_TYPE_SIZE > INT_TYPE_SIZE)))
      spit_string("long long");

    else if (type == long_integer_type_node
        || type == long_unsigned_type_node
        || (TYPE_PRECISION(type) == LONG_TYPE_SIZE
        && (LONG_TYPE_SIZE > INT_TYPE_SIZE)))
      spit_string("long");

    else if (type == char_type_node
        || type == unsigned_char_type_node
      || TYPE_PRECISION(type) == CHAR_TYPE_SIZE)
      spit_string("char");

    else if (type == signed_char_type_node)
      spit_string("signed char");

    else if (type == integer_type_node ||
      type == unsigned_type_node)
      spit_string("int");

    else
      spit_string("/* ?INTEGER_TYPE */ int");
    break;

  case REAL_TYPE:
    spit_type_qual(qualified_type);

    if (type == float_type_node
      || (TYPE_PRECISION(type) == FLOAT_TYPE_SIZE
        && (FLOAT_TYPE_SIZE < DOUBLE_TYPE_SIZE)))
      spit_string("float");

    else if (type == long_double_type_node
        || (TYPE_PRECISION(type) == LONG_DOUBLE_TYPE_SIZE
        && (LONG_DOUBLE_TYPE_SIZE > DOUBLE_TYPE_SIZE)))
      spit_string("long double");

    else if (type == double_type_node)
      spit_string("double");

    else
      spit_string("/* ?REAL_TYPE */ double");

    break;

  case RECORD_TYPE:
  case UNION_TYPE:
    spit_type_qual(qualified_type);

    spit_record_type(type);
    break;

  case ENUMERAL_TYPE:
    spit_type_qual(qualified_type);

    spit_enum_type(type);
    break;

  case ARRAY_TYPE:
    spit_string("/* ARRAY_TYPE is not a basic type */ ");
    break;

  case POINTER_TYPE:
    spit_string("/* POINTER_TYPE is not a basic type */ ");
    break;

  case FUNCTION_TYPE:
    spit_string("/* FUNCTION_TYPE is not a basic type */ ");
    break;

  default:
    spit_string("/* Unknown basic type */ ");
  }
}

void
spit_type_basename(tree decl)
{
  tree  basic_type;

  if (decl != NULL_TREE)
  {
    basic_type = get_base_type(TREE_TYPE(decl));

    /* This case handles type trees whose bottom is a TYPE_DECL. This kind of
       tree is not produced by the front end, but it is introduced by some
       analyses, e.g. EARTH-C analysis. */

    if (basic_type != NULL_TREE
      && TREE_CODE(basic_type) == TYPE_DECL)
      spit_name(basic_type);

    /* If we can remember the typedef name that was used in the declaration
       of the variable, use that as the base type. For example, if we have
       the following in a C program:
    
    typedef int *My_type; My_type var;
    
    we prefer dumping
    
    My_type var;
    
    to dumping
    
    int *var;
    
    The previous version of the c-dump.c file did this to some extent, but
       allowed only one typedef name per type. For example, the C code
    
    typedef int T1; typedef int T2;
    
    would cause all int variables to be output as T1. */

    else if (DECL_TYPE_NAME(decl) != NULL_TREE
      && TREE_CODE(DECL_TYPE_NAME(decl)) == TYPE_DECL)
    {
      if (TREE_CODE(decl) != FIELD_DECL)
      {
        spit_type_qual(basic_type);
        spit_name(DECL_TYPE_NAME(decl));
      }
      else
        spit_base_type(basic_type);
    }
    else
    {
      spit_base_type(basic_type);
    }
  }
}

/***************************************************************************

Search the type for array sizes, function parameter lists, etc.

***************************************************************************/

tree
get_nth_type_constr(tree type, int n)
{
  if (n == 0)
    return type;

  else
  {
    switch (TREE_CODE(type))
    {
    case ARRAY_TYPE:
    case POINTER_TYPE:
      return get_nth_type_constr(TREE_TYPE(type), n - 1);
      break;

    case FUNCTION_TYPE:
      return get_nth_type_constr(TREE_TYPE(type), n - 1);
      break;

    default:
      return type;
    }
  }
}

static void
spit_arg_types(tree arg_types)
{
  while (arg_types != NULL_TREE
    && TREE_CODE(arg_types) == TREE_LIST
    && TREE_VALUE(arg_types) != NULL_TREE
    && TREE_CHAIN(arg_types) != NULL_TREE)
  {
    tree  value = TREE_VALUE(arg_types);

    spit_base_type(get_base_type(value));
    new_spit_type_i(value, build_tc_from_type(value),
      NULL_TREE, tc_none, get_tc_length_from_type(value));

    arg_types = TREE_CHAIN(arg_types);

    if (TREE_CHAIN(arg_types) != NULL_TREE)
      spit_string(", ");
  }
}

/***************************************************************************

new_spit_type_i() regenerates the type from the type constructor type_constr,
which is a string a pairs of bits, each of which represents a type
constructor: 00 = none, 01 = ptr, 10 = array, 11 = function.  The rightmost
bit pair contains the outermost constructor, i.e. the one of lowest
precedence.  This format has no space for more precise information, such as
array sizes and function parameter lists, but we recover these from the type
itself.

The parameter `prev_constr' is used to handle precedence of constructors
(for proper parenthesization), e.g. `[]' has higher precedence than `*'.

***************************************************************************/

static void
new_spit_type_i(tree type, unsigned long type_constr,
  tree var_name, int prev_constr, int depth)
{
  int   cur_tc = type_constr & 3;
  tree  more_info;
  tree  basic_type;

  if (type != NULL_TREE && TREE_CODE(type) == TYPE_DECL)
  {
    spit_name(type);
    spit_string(" ");
    return;
  }
  switch (cur_tc)
  {
  case tc_none:
    if (var_name != NULL_TREE)
    {
      if ((TREE_CODE(var_name) == FUNCTION_DECL) &&
          (!(IS_PSEUDO_FUNCTION(var_name) || TREE_EXTERNAL(var_name))))
      {
	if (dump_func_prototype == 1)
	{ /* dumping the prototype so set up a link to the actual func */
	  spit_html_link_string(c_file, var_name, 
		IDENTIFIER_POINTER(DECL_NAME(var_name)), 
                c_base_name, ".html", "source");
	}
	else
	{
	  spit_html_target_string(c_file, var_name, NULL);
	  spit_html_cgi_link_string(c_file, var_name, NULL, NULL, NULL);
          spit_string(IDENTIFIER_POINTER(DECL_NAME(var_name)));
          spit_html_string("</FONT></a>");
	}
      }
#if 0 /* Rakesh : don't want to do this any more: Rakesh Jan 23/96 */
      else if ((TREE_CODE(var_name) == PARM_DECL) && (dump_cur_func) &&
               (!(IS_PSEUDO_FUNCTION(dump_cur_func) || 
         	  TREE_EXTERNAL(dump_cur_func))))
      {
        if (HEAP_STUFF(dump_cur_func) && HEAP_FLAG(dump_cur_func) && 
            INDEX_INFO_INDEX(heap_get_index(var_name, dump_cur_func)) >= 0) 
        /* only if parm is a heap-directed ptr */
	  spit_html_link_string(c_file, var_name, 
            IDENTIFIER_POINTER(DECL_NAME(var_name)), c_base_name, 
            ".give.info.html", NULL);
        else
          spit_string(IDENTIFIER_POINTER(DECL_NAME(var_name)));
      }
#endif
      else if (TREE_CODE(var_name) == FIELD_DECL) 
      {
	spit_html_target_string(c_file, var_name, NULL);
	if (current_struct_node != NULL)
	  spit_html_link_string(c_file, current_struct_node, 
	    IDENTIFIER_POINTER(DECL_NAME(var_name)), NULL, 
	    NULL, NULL);
      }
      else if (TREE_CODE(var_name) == TYPE_DECL)
      {
	if (dump_typedefs == 1)
	  spit_html_target_string(c_file, var_name, 
	   IDENTIFIER_POINTER(DECL_NAME(var_name)));
        else
	  spit_html_link_string(c_file, var_name, IDENTIFIER_POINTER(
	     DECL_NAME(var_name)), c_base_name, ".typedefs.html", NULL);
      }
      else 
      {
       /* Rakesh : Dec 10/96: now var_name is actually the DECL node itself */
       spit_string(IDENTIFIER_POINTER(DECL_NAME(var_name)));
      }
    }
    break;

  case tc_ptr:
    /* If the tree hierarchy is opposite of precedence, e.g. if a function
       type or array type is the parent of a pointer type, spit parentheses
       to protect this precedence. */
    if (prev_constr == tc_func || prev_constr == tc_array)
      spit_string("(");

    spit_string("*");

    more_info = get_nth_type_constr(type, depth - 1);

    if (more_info != NULL_TREE && TREE_CODE(more_info) == POINTER_TYPE)
      spit_type_qual(more_info);

    new_spit_type_i(type, type_constr >> 2, var_name, cur_tc, depth - 1);

    if (prev_constr == tc_func || prev_constr == tc_array)
      spit_string(")");
    break;

  case tc_array:
    new_spit_type_i(type, type_constr >> 2, var_name, cur_tc, depth - 1);
    spit_string("[");
    more_info = get_nth_type_constr(type, depth - 1);

    if (more_info == NULL_TREE
      || TREE_CODE(more_info) != ARRAY_TYPE)
      spit_string("/*??*/");
    else
      spit_array_size(more_info);

    spit_string("]");
    break;

  case tc_func:
    new_spit_type_i(type, type_constr >> 2, var_name, cur_tc, depth - 1);
    more_info = get_nth_type_constr(type, depth - 1);

    spit_string("(");
    save_column = current_column;

    if (more_info == NULL_TREE
      || TREE_CODE(more_info) != FUNCTION_TYPE)
      spit_string("/* ?args */");
    else
    {
      if (TREE_CODE(type) != FUNCTION_TYPE)
      {
        spit_arg_types(TYPE_ARG_TYPES(more_info));
      }
      else
      {
        /* Rakesh : Nov 17/96: Don't dump psuedo-codes from prototypes.c */
        if (!IS_PSEUDO_FUNCTION(dump_cur_func))
          spit_params(TYPE_ARG_LIST(more_info));
      }
    }
    spit_string(")");
    break;
  }
}

static void
spit_decorated_name(tree decl)
{
  tree  base_type, decl_type, typedef_name, decl_name;
  unsigned long type_constr;
  int   tc_length_typedef, tc_length_all, tc_length;
  int   code;
  char  buf[80];

  if (decl == NULL_TREE)
    return;

  code = TREE_CODE(decl);
  decl_type = TREE_TYPE(decl);
  base_type = get_base_type(decl_type);
  typedef_name = DECL_TYPE_NAME(decl);
  tc_length_all = get_tc_length_from_type(decl_type);
  decl_name = DECL_NAME(decl);

  /* The inner parts of the type may be implied by the typedef name in the
     declaration.  If so, we determine how much of the type constructor is
     actually implicit, in order that we not dump it explicitly.  For
     example, consider the declarations
  
  typedef int A10[10]; A10 *a;
  
  `a' has type `pointer to an array of 10 integers', but we do not want to
     dump the array constructor, because it is implicit in the typedef type
     of `a'. */
  if (typedef_name && TREE_CODE(base_type) != TYPE_DECL)
    tc_length_typedef = get_tc_length_from_type(TREE_TYPE(typedef_name));
  else
    tc_length_typedef = 0;

  tc_length = tc_length_all - tc_length_typedef;

  /* Naively build the type constructor which represents the whole type
     explicitly, and then discard the part (the righmost bits) which is
     implicit from the typedef type. */
  type_constr = build_tc_from_type(decl_type) >> (2 * tc_length_typedef);

#if 0
  sprintf(buf, " /* %s %06lx */ \0",
    typedef_name != NULL_TREE ?
    IDENTIFIER_POINTER(DECL_NAME(typedef_name)) : "", type_constr);
  print_file_or_string(buf);
#endif

  /* Rakesh : pass decl node itself, instead of decl_name: no harm and
   * gives me the info in new_spit_type_i if it is a function decl when
   * the name is being printed i: Dec 10, 96*/
  /* new_spit_type_i(decl_type, type_constr, decl_name, tc_none, tc_length); */
  new_spit_type_i(decl_type, type_constr, decl, tc_none, tc_length);
}

EXPORT void
spit_decl(tree decl, int indent_flag, int spit_base_type_flag)
{
  enum tree_code decl_code = TREE_CODE(decl);

  if (indent_flag == INDENT)
    spit_string(indent_str);

  switch (decl_code)
  {
  case TYPE_DECL:
    spit_string("typedef ");
    spit_type_basename(decl);
    spit_string(" ");
    spit_decorated_name(decl);
    break;

  case VAR_DECL:
    if (spit_base_type_flag)
      spit_storage_class(decl);

    if (spit_base_type_flag)
    {
      spit_type_basename(decl);
      spit_string(" ");
    }
    spit_decorated_name(decl);

    if (EARTH_dump || dump_EARTH_comments)
      if (DECL_PROC_NO(decl) != NULL_TREE)
      {
        if (dump_EARTH_comments)
          spit_string(" /* ");

        spit_string(" @ ");
        spit_cst(DECL_PROC_NO(decl));

        if (dump_EARTH_comments)
          spit_string(" */ ");
      }
    break;

  case PARM_DECL:
    spit_type_basename(decl);
    spit_string(" ");
    spit_decorated_name(decl);
    break;

  case FIELD_DECL:
    spit_type_basename(decl);
    spit_string(" ");
  /* Rakesh : pass decl node itself, instead of decl_name: no harm and
   * gives me the info in new_spit_type_i if it is a function decl when
   * the name is being printed i: Dec 10, 96*/
    new_spit_type_i(TREE_TYPE(decl),
      build_tc_from_type(TREE_TYPE(decl)),
      decl, tc_none,
      get_tc_length_from_type(TREE_TYPE(decl)));
    /* 
    new_spit_type_i(TREE_TYPE(decl),
      build_tc_from_type(TREE_TYPE(decl)),
      DECL_NAME(decl), tc_none,
      get_tc_length_from_type(TREE_TYPE(decl)));
      */
#if 0
    spit_decorated_name(decl);
#endif
    if (basic_type(decl))
      spit_bit_field_size(decl);
    break;

  case FUNCTION_DECL:
    spit_type_basename(decl);
    spit_string(" ");
    spit_decorated_name(decl);
    break;

  case RESULT_DECL:
    spit_string("/* RESULT_DECL */");
    break;

  default:
    fprintf(stderr, "C-DUMP: spit_decl(): Unknown declaration type = %s.\n",
      tree_code_name[(int) TREE_CODE(decl)]);
  }
}

/* This function is intended as a replacement to the old `spit_type()', i.e.,
   it can be used in the same manner.  It dumps the full type explicitly,
   that is, it knows nothing about the typedefs that may have been used in
   the declaration which contains this type. */
void
new_spit_type(tree type)
{
  tree  base_type;
  unsigned long type_constr;
  int   tc_length;
  /* char                buf[80]; */

  if (type == NULL_TREE)
    return;

  base_type = get_base_type(type);
  tc_length = get_tc_length_from_type(type);
  type_constr = build_tc_from_type(type);

  spit_base_type(base_type);
  spit_string(" ");
  new_spit_type_i(type, type_constr, NULL_TREE, tc_none, tc_length);
}

LOCAL void
spit_array_size(tree array_type)
{
  if (array_type == NULL_TREE || TREE_CODE(array_type) != ARRAY_TYPE)
    spit_string("/*?array size*/");
  else
  {
    char  str[16];
    long  size;

    if (TYPE_DOMAIN(array_type) && TYPE_MAX_VALUE(TYPE_DOMAIN(array_type)))
      size = 1 + TREE_INT_CST_LOW(TYPE_MAX_VALUE(TYPE_DOMAIN(array_type)));
    else
      size = -1;

    sprintf(str, "%ld", size);

    if (size > 0)
      spit_string(str);
  }
}

static void
spit_type_cast(tree type_cast)
{
  tree  type;

  if (type_cast == NULL_TREE)
    return;

  if (TREE_CODE(type_cast) != NOP_EXPR
    && TREE_CODE(type_cast) != CONVERT_EXPR)
    return;

  type = TREE_TYPE(type_cast);

  spit_string("(");

  spit_base_type(get_base_type(type));
  spit_string(" ");
  new_spit_type_i(type, build_tc_from_type(type),
    NULL_TREE, tc_none, get_tc_length_from_type(type));

  spit_string(")");
}

static int
get_tc_length_from_type(tree type)
{
  if (type == NULL_TREE)
    return 0;

  switch (TREE_CODE(type))
  {
  case VOID_TYPE:
  case INTEGER_TYPE:
  case REAL_TYPE:
  case RECORD_TYPE:
  case UNION_TYPE:
  case ENUMERAL_TYPE:
    return 0;
    break;

  case TYPE_DECL:
    return 0;
    break;

  case POINTER_TYPE:
  case ARRAY_TYPE:
  case FUNCTION_TYPE:
    return (1 + get_tc_length_from_type(TREE_TYPE(type)));
    break;

  default:
    return 0;
  }
}

static unsigned long
build_tc_from_type(tree type)
{
  unsigned long acc = 0;

  if (type == NULL_TREE)
    return 0;

  while (TREE_CODE(type) == POINTER_TYPE
    || TREE_CODE(type) == ARRAY_TYPE
    || TREE_CODE(type) == FUNCTION_TYPE)
  {
    acc <<= 2;

    if (TREE_CODE(type) == POINTER_TYPE)
      acc |= tc_ptr;

    else if (TREE_CODE(type) == ARRAY_TYPE)
      acc |= tc_array;

    else
      acc |= tc_func;

    type = TREE_TYPE(type);
  }

  return acc;
}

/************************************************************************
 *
 * spit_SSA_Number()            Added by C. Lapkowski
 * This function prints out the SSA number beside the variable
 *
 ************************************************************************/
static void
spit_SSA_Number(tree decl, tree stmt, int side)
{
  char  number[20];
  int   num, ext_number;
  extern int dump_SSA_Numbers_flag;

  if (stmt && dump_SSA_Numbers_flag)
  {
    num = get_SSA_number(decl, stmt, side, &ext_number);
#if 0 /* only the primary ssa num please */
    if (TREE_CODE(TREE_TYPE(decl)) == POINTER_TYPE ||
      (TREE_CODE(TREE_TYPE(decl)) == ARRAY_TYPE &&
        TREE_CODE(decl) == PARM_DECL))
      sprintf(number, "%1u,%1u", num, ext_number);
    else
#endif
      sprintf(number, "%1u", num);

#if 0
    spit_string("/*"); /* Rakesh*/
#endif
 if (TREE_CODE(decl) != COMPONENT_REF && num != 0)
 {
    spit_html_string("<SUB><FONT color=tomato>");
    /* print_file_or_string(number); */
    spit_html_string(number); 
    spit_html_string("</FONT></SUB>");
 }
#if 0
    spit_string("*/");
#endif
  }
}


/* Rakesh Nov 26/96: check if the parent of an ADDR_EXPR is array ref */
LOCAL int
parent_is_array_ref(addr_expr_node, stmt, side)
  tree  addr_expr_node;
  tree  stmt;
  int   side;
{
  tree  s_body;

  if (side == SSA_LHS)
   return 1; /* no need to dump the addr operator on lhs anyway */
  /* We return 1, so that height_difference check is not made for ADDR_EXPR*/
  /* Else it aborts */
  if (!stmt)
    return 1;                   /* empty stmt: possible during calls to
                                   spit_expr from simplify
                                   (get_comment_string ) */
  if (TREE_CODE(stmt) == EXPR_STMT)
    s_body = STMT_BODY(stmt);
  else if (TREE_CODE(stmt) == MODIFY_EXPR)
    s_body = stmt;
  else
    return 0;                   /* stmt cannot contain an array ref otherwise */
  if (TREE_CODE(s_body) != MODIFY_EXPR)
    return 0;                   /* stmt cannot contain an array ref otherwise */
  if (side == SSA_LHS)
    side = 0;
  else if (side == SSA_RHS)
    side = 1;
  else
    abort();                    /* something wrong */
  /* if the RHS operand of the modify expr is the given add_expr_node,
   * it implies the rhs ref is of type lhs = &(arr) and that the "&" operator
   * is not a child of an ARRAY_REF node */
  if ((TREE_OPERAND(s_body, side)) &&
      (TREE_OPERAND(s_body, side) == addr_expr_node))
    return 0;
  return 1;
}

/* Rakesh Nov 29/96: compare the # of TYPE nodes in 
 * the type chain of an ADDR_EXPR node and its operand (of
 * ARRAY_TYPE) : return 0 if 
 * the number is same, 1 if difference is 1, else abort: 
 should not happen, so */
LOCAL int
height_difference(tree addr_type, tree op_type)
{
  int addr_len, op_len;

  addr_len = 0; op_len = 0;
  while ((addr_type))
  {
   addr_len++;
   addr_type = TREE_TYPE(addr_type);
  }
  while (op_type)
  {
   op_len++;
   op_type = TREE_TYPE(op_type);
  }
  if (op_len == addr_len)
   return 0;
  /*else */ if (op_len == (addr_len - 1))
   return 1;
   /*else*/
   abort(); /* something wrong */
}

LOCAL void
spit_all_frames()
{
 FILE *t_file;
 t_file = frames_dump_file;
 fprintf(t_file, "<HTML>\n");
 fprintf(t_file, "<FRAMESET rows=\"5\%,5\%,90\%\">\n");
 fprintf(t_file, "<FRAMESET>\n");
 fprintf(t_file, "<FRAME name=\"menu\" src=\"%s.menu.html\">\n", c_base_name);
 fprintf(t_file, "<FRAME name=\"note\" src=\"%s.note.html\"\n", c_base_name);
 fprintf(t_file, "</FRAMESET>\n");
 fprintf(t_file, "<FRAMESET cols=\"60\%,40\%\">\n");
 fprintf(t_file, "<FRAME name=\"source\" src=");
 fprintf(t_file, "\"%s.html\" scrolling=\"auto\">\n", c_base_name);
 fprintf(t_file,"<FRAME name=\"info\" src=\"%s.info.html\" scrolling=\"auto\">\n", c_base_name);
 fprintf(t_file, "</FRAMESET>\n");
 fprintf(t_file, "</FRAMESET>\n");
 fprintf(t_file, "</HTML>\n");
 fclose(t_file);
 fprintf(info_dump_file, "<HTML>\n");
 fprintf(info_dump_file, "This frame will display flow information\n");
 fprintf(info_dump_file, "</HTML>\n");
 fclose(info_dump_file);
 fprintf(note_dump_file, "<HTML>\n");
 fprintf(note_dump_file, "<BODY bgcolor=white>\n");
 fprintf(note_dump_file, "Please click at the information you want to browse.");
 fprintf(note_dump_file, " Currently <FONT color=red>no information</FONT>");
 fprintf(note_dump_file, " is being viewed.\n");
 fprintf(note_dump_file, "</HTML>\n");
 fclose(note_dump_file);
}

LOCAL void
spit_all_menus()
{
 FILE *t_file;
 char *ch_buffer, *temp, *lastslash, *output_dir_name;
 char *path_name;
 path_name = full_path_name;
 t_file = menu_dump_file;
 fprintf(t_file, "<HTML>\n");
 fprintf(t_file, "<BODY bgcolor=white>\n");
 fprintf(t_file, "<TABLE>\n");
 fprintf(t_file, "<TD>\n");
 fprintf(t_file, "<BASE HREF=\"http://www-acaps.cs.mcgill.ca/\" ");
 fprintf(t_file, "TARGET = \"note\">\n");
 if (dump_read_write_sets) 
 {
   fprintf(t_file, "<A HREF = ");
   fprintf(t_file, 
     "\"cgi-bin/ghiya/flow-info.pl?info=rw_sets");
   fprintf(t_file, "&file=%s/%s", path_name, c_base_name);
   fprintf(t_file, "\"><FONT color=red>[Stack R/W Sets]</FONT></a>\n");
 }
 if (flag_connect_rw_dump) 
 {
   fprintf(t_file, "<A HREF = ");
   fprintf(t_file, 
     "\"cgi-bin/ghiya/flow-info.pl?info=heap_rw_sets");
   fprintf(t_file, "&file=%s/%s", path_name, c_base_name);
   fprintf(t_file, "\"><FONT color=red>[Heap R/W Sets]</FONT></a>\n");
 }
 if (flag_print_fields) 
 {
   fprintf(t_file, "<A HREF = ");
   fprintf(t_file, 
     "\"cgi-bin/ghiya/flow-info.pl?info=fields");
   fprintf(t_file, "&file=%s/%s", path_name, c_base_name);
   fprintf(t_file, "\"><FONT color=red>[Field-Sets]</FONT></a>\n");
 }
 if (flag_connect_indirect_dump)
 {
   fprintf(t_file, "<A HREF = ");
   fprintf(t_file, 
     "\"cgi-bin/ghiya/flow-info.pl?info=connect_indirect");
   fprintf(t_file, "&file=%s/%s", path_name, c_base_name);
   fprintf(t_file, "\"><FONT color=red>[Connection]</FONT></a>\n");
 }
 if (flag_connect_dump)
 {
   fprintf(t_file, "<A HREF = ");
   fprintf(t_file, 
     "\"cgi-bin/ghiya/flow-info.pl?info=connect");
   fprintf(t_file, "&file=%s/%s", path_name, c_base_name);
   fprintf(t_file, "\"><FONT color=red>[All-Connection]</FONT></a>\n");
 }
 if (Alias_indirect_dump)
 {
   fprintf(t_file, "<A HREF = ");
   fprintf(t_file, 
     "\"cgi-bin/ghiya/flow-info.pl?info=alias_indirect");
   fprintf(t_file, "&file=%s/%s", path_name, c_base_name);
   fprintf(t_file, "\"><FONT color=red>[Points-to]</FONT></a>\n");
 }
 if (Alias_dump)
 {
   fprintf(t_file, "<A HREF = ");
   fprintf(t_file, 
     "\"cgi-bin/ghiya/flow-info.pl?info=alias");
   fprintf(t_file, "&file=%s/%s", path_name, c_base_name);
   fprintf(t_file, "\"><FONT color=red>[All-Points-to]</FONT></a>\n");
 }
 fprintf(t_file, "<A HREF = ");
 fprintf(t_file,
     "\"~ghiya/info.html\" TARGET=\"info\"");
 fprintf(t_file, "\"><FONT color=red>[Show Intro]</FONT></a>\n");

 fprintf(t_file, "</TD>\n");
 fprintf(t_file, "</TABLE>\n");
 fprintf(t_file, "</HTML>\n");
 fclose(t_file);
 temp = ast_c_dump_base_name;
 output_dir_name = (char *) Mmalloc(strlen(ast_c_dump_base_name) + 8);
 ch_buffer = (char *) Mmalloc(strlen(ast_c_dump_base_name) + 40);
 while (*temp != '\0')
 {
    if (*temp == '/')
    {
      lastslash = temp;
      /* the name of the current directory in the path */
      memset(output_dir_name, 0, strlen(ast_c_dump_base_name) + 8);
      strncpy(output_dir_name, ast_c_dump_base_name,
          ((int) lastslash - (int) ast_c_dump_base_name));
      sprintf(ch_buffer, "chmod a+rx %s", output_dir_name);
      system(ch_buffer);
    }
    temp++;
 }
 sprintf(ch_buffer, "chmod a+r %s.*html", ast_c_dump_base_name);
 system(ch_buffer);

}

LOCAL void
spit_call_expr_num(tree call_expr_node)
{
  char number[20];
  int num;
  extern int dump_SSA_Numbers_flag;
  if (!dump_SSA_Numbers_flag)
   return;
  num = CALL_EXPR_NUM(call_expr_node);
  if (num == 0)
   return; /* external call */
  sprintf(number, "%1u", num);
  spit_html_string("<FONT color=red>");  
  spit_html_string("<SUB>");  
  spit_string(number);
  spit_html_string("</SUB>");  
  spit_html_string("</FONT>");  

}
/******************** END OF C DUMP ************************************/
/***********************************************************************/
