/**
   Copyright (C) 1993 McGill University.
   Copyright (C) 1993 McCAT System Group.
   Copyright (C) 1996 Christopher Lapkowski

This program is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 1, or (at your option) any later
   version.

This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 675
   Mass Ave, Cambridge, MA 02139, USA.
**/
/**************************************************************************
 *
 * MODULE: dh.c
 *
 *      This module implements some general purpose dump hooks for the
 *      C-Dump.
 *
 **************************************************************************/
static char rcsid[] = "$Id: dh.c,v 1.3 1996/11/26 23:20:45 mcadmin Exp $";

#include <stdio.h>
#include "tree.h"
#include "al.external.h"
#include "analysis.h"
#include "heap.external.h"
#include "dh.h"

int   dump_uid(tree, int);
int   dump_stmt_alias(tree, int);
int   dump_stmt_eqc(tree, int);
int   dump_stmt_indirect_alias(tree, int);
int   dump_after_func_alias(tree, int);
int   dump_read_write(tree, int);
int   dump_init_dam_statistics(tree, int);
int   dump_dam_statistics(tree, int);
int   dump_dam(tree, int);
int   dump_dam_expr(tree, int);
int   dump_dam_indent(tree, int);
int   dump_dam_unindent(tree, int);
int   dump_stmt_connect(tree, int);
int   dump_stmt_indirect_connect(tree, int);
int   dump_stmt_rw_connect(tree, int);
int   dump_heap_loop_invar(tree, int);
int   dump_heap_loc_invar(tree, int);
int   dump_heap_cse(tree, int);
int   dump_heap_print_fields(tree, int);

extern tree dump_cur_func;      /* c-dump.c */
extern FILE *connect_indirect_dump_file; /* c-dump.c */
extern FILE *connect_dump_file; /* c-dump.c */
extern FILE *alias_indirect_dump_file; /* c-dump.c */
extern FILE *alias_dump_file; /* c-dump.c */
extern FILE *fields_dump_file; /* c-dump.c */
extern FILE *heap_rw_dump_file; /* c-dump.c */
FILE *c_file;

extern int dump_UID_flag;
extern int dump_read_write_sets;
extern int Alias_dump;          /* the flag to print the result of alias
                                   process */
extern int Alias_fast_dump;
extern int flag_connect_dump;
extern int flag_connect_indirect_dump;
extern int flag_connect_rw_dump;
extern int flag_heap_loop_invars;
extern int flag_heap_cse;
extern int Alias_indirect_dump; /* the flag to print the alias info only for
                                   indirect refs in the program */
extern int flag_print_fields; /* prints fields accessed wrt heap pointers */
extern int dam_spit_dep;        /* Flags for McADT */
extern int dam_statistic;
FILE *damstat_file;
char *damstat_name;

extern char *ast_c_dump_base_name;

void
setup_dump_hooks()
{
  if (dump_UID_flag)
  {
    dh_insert(dump_uid, DH_BeforeStmt | DH_BeforeFunc | DH_BeforeCond | DH_BeforeInit | DH_BeforeIter);
  }
  if (Alias_dump)
    /* Rakesh, Oct 7/96: removed the Alias_indirect_dump flag : with this
       flag we only want the alias info for indirect refs */
  {
    dh_insert(dump_stmt_alias, DH_BeforeStmt);
    dh_insert(dump_after_func_alias, DH_AfterFunc);
  }
  if (Alias_indirect_dump)
    /* Rakesh, Oct 7/96: added to only dump alias info at indirect refs */
  {
    dh_insert(dump_stmt_indirect_alias, DH_BeforeStmt);
    /* Rakesh: we can have indirect refs in init and iter stmts for FOR_STMT */
    dh_insert(dump_stmt_indirect_alias, DH_BeforeInit);
    dh_insert(dump_stmt_indirect_alias, DH_BeforeIter);
  }
  if (Alias_fast_dump){
    dh_insert(dump_stmt_eqc, DH_BeforeStmt);
  }
  if (flag_connect_dump)
    /* Rakesh, Nov 12/96: added to dump connection info at each stmt */
  {
    dh_insert(dump_stmt_connect, DH_BeforeStmt);
    dh_insert(dump_stmt_connect, DH_AfterFunc);
  }
  if (flag_connect_indirect_dump)
    /* Rakesh, Nov 12/96: dump connection info only at heap indirect refs */
  {
    dh_insert(dump_stmt_indirect_connect, DH_BeforeStmt);
  }
  if (flag_connect_rw_dump)
    /* Rakesh, Nov 12/96: dump heap read-write info */
  {
    dh_insert(dump_stmt_rw_connect, DH_BeforeStmt);
    dh_insert(dump_stmt_rw_connect, DH_BeforeFunc);
  }
  /* Rakesh, Nov 12/96: indicate stmts involved in heap loop invar motion */
  if (flag_heap_loop_invars)
  {
    dh_insert(dump_heap_loop_invar, DH_BeforeStmt);
  }
  /* Rakesh, Nov 12/96: indicate stmts involved in heap CSE */
  if (flag_heap_cse)
  {
    dh_insert(dump_heap_cse, DH_BeforeStmt);
  }
  /* Rakesh, Jan 23/97: dump fields accessed wrt heap pointers */
  if (flag_print_fields)
  {
    dh_insert(dump_heap_print_fields, DH_BeforeStmt);
    dh_insert(dump_heap_print_fields, DH_BeforeFunc);
  }
  if (dump_read_write_sets)
  {
    dh_insert(dump_read_write, DH_BeforeStmt | DH_BeforeFunc);
  }
  if (dam_spit_dep)
  {
    dh_insert(dump_dam, DH_AfterStmt);
    dh_insert(dump_dam_expr, DH_BeforeStmt);
    dh_insert(dump_dam_indent, DH_BeforeFunc | DH_AfterStmt);
    dh_insert(dump_dam_unindent, DH_AfterFunc);
    if (dam_statistic)
    {
      dh_insert(dump_init_dam_statistics, DH_Before);
      dh_insert(dump_dam_statistics, DH_After);
    }
  }
  gcp_revdump();

}

void
unset_dump_hooks()
{
  if (dump_UID_flag)
  {
    dh_remove(dump_uid, DH_BeforeStmt | DH_BeforeFunc | DH_BeforeCond | DH_BeforeInit | DH_BeforeIter);
  }
  if (Alias_dump)
  {
    dh_remove(dump_stmt_alias, DH_BeforeStmt);
    dh_remove(dump_after_func_alias, DH_AfterFunc);
  }
  if (Alias_indirect_dump)
  {
    dh_remove(dump_stmt_indirect_alias, DH_BeforeStmt);
    dh_remove(dump_stmt_indirect_alias, DH_BeforeInit);
    dh_remove(dump_stmt_indirect_alias, DH_BeforeIter);
  }
  if (Alias_fast_dump){
    dh_remove(dump_stmt_eqc, DH_BeforeStmt);
  }

  if (flag_connect_dump)
  {
    dh_remove(dump_stmt_connect, DH_BeforeStmt);
  }
  if (flag_connect_indirect_dump)
  {
    dh_remove(dump_stmt_indirect_connect, DH_BeforeStmt);
  }
  if (flag_connect_rw_dump)
  {
    dh_remove(dump_stmt_rw_connect, DH_BeforeStmt);
    dh_remove(dump_stmt_rw_connect, DH_BeforeFunc);
  }
  if (flag_heap_loop_invars)
  {
    dh_remove(dump_heap_loop_invar, DH_BeforeStmt);
  }
  if (flag_heap_cse)
  {
    dh_remove(dump_heap_cse, DH_BeforeStmt);
  }

  /* Rakesh, Jan 23/97: dump fields accessed wrt heap pointers */
  if (flag_print_fields)
  {
    dh_remove(dump_heap_print_fields, DH_BeforeStmt);
    dh_remove(dump_heap_print_fields, DH_BeforeFunc);
  }

  if (dump_read_write_sets)
  {
    dh_remove(dump_read_write, DH_BeforeStmt | DH_BeforeFunc);
  }
  if (dam_spit_dep)
  {
    dh_remove(dump_dam, DH_AfterStmt);
    dh_remove(dump_dam_expr, DH_BeforeStmt);
    dh_remove(dump_dam_indent, DH_BeforeFunc);
    dh_remove(dump_dam_unindent, DH_AfterFunc);
    if (dam_statistic)
    {
      dh_remove(dump_init_dam_statistics, DH_Before);
      dh_remove(dump_dam_statistics, DH_After);
    }
  }
  gcp_unrevdump();
}

/**********************************************************************
 *              USER DUMP FUNCTIONS
 **********************************************************************/

/********************** TREE_UID DUMP *********************************/
int
dump_uid(tree node, int indent)
{
  char  buf[10];

  if (node && TREE_CODE(node) != TREE_LIST)
  {
    if (TREE_CODE(node) == AST_DECL_NODE)
      sprintf(buf, "/* %d */", TREE_UID(AST_DECL_PTR(node)));
    else
      sprintf(buf, "/* %d */", TREE_UID(node));
    spit_string(buf);
  }
}

/********************** POINTS-TO DUMP *********************************/

int
dump_stmt_alias(tree node, int indent)
{
  /*------------ MAR -------------*/
  if (TREE_CODE(node) != TREE_LIST)
  {
    /*--- second parameter is the indentation length */
    al_print_info_html(alias_dump_file, node, indent);
  }
}

/* Rakesh, Oct 7/96: created to only dump alias info at indirect refs */
int
dump_stmt_indirect_alias(tree node, int indent)
{
  if (node == NULL)
    return;                     /* Possible as FOR_STMT can have empty init
                                   and iter stmts */

  if (TREE_CODE(node) == MODIFY_EXPR)   /* iter or init stmt */
    al_indirect_print_info_html(alias_indirect_dump_file, node, indent);

  /* Rakesh: Prints points-to info only for indirect refs in the program */
  if (TREE_CODE(node) == EXPR_STMT)
  {
    if ((TREE_CODE(STMT_BODY(node)) == MODIFY_EXPR) ||
      (TREE_CODE(STMT_BODY(node)) == CALL_EXPR))
    {
      al_indirect_print_info_html(alias_indirect_dump_file,  node, indent);
    }
  }

}

int
dump_after_func_alias(tree node, int indent)
{
  al_print_info_html(alias_dump_file, AST_DECL_PTR(node), 0);
}

/********************** READ/WRITE SETS DUMP *****************************/

int
dump_read_write(tree node, int indent)
{
  if (!IS_IN_CALL_GRAPH(dump_cur_func))
    return;                     /* No rw info for these functions: Rakesh:
                                   Nov 19/96 */
  if (TREE_CODE(node) == AST_DECL_NODE)
    rw_print_global_info(AST_DECL_PTR(node), 0);
  else
    rw_print_info(node, indent);
}

/********************** McADT DUMP *********************************/

int
dump_init_dam_statistics(tree node, int indent)
{
  register char *dumpname;
  dumpname = (char *) Mmalloc(strlen(ast_c_dump_base_name) + 20);

  sprintf(dumpname, "%s.damstat", ast_c_dump_base_name);
  damstat_name = dumpname;
  /**MYDEBUG**/
  fprintf(stderr, "damstat_file name = %s\n", dumpname);

  damstat_file = fopen(dumpname, "w");
  if (damstat_file == 0)
    pfatal_with_name(dumpname);
}

int
dump_dam_statistics(tree node, int indent)
{
  FILE *temp;

  if (mydebug)
    fprintf(stderr, "STARTING TO DUMP DAMSTAT \n");

  temp = c_file;
  c_file = damstat_file;
  dam_spit_statistic();
  fclose(c_file);
  c_file = temp;
}

int
dump_dam_indent(tree node, int indent)
{
  if (TREE_CODE(node) == AST_DECL_NODE || TREE_CODE(node) == EXPR_STMT)
  {
    inc_indent();
    inc_indent();
    inc_indent();
  }
}

int
dump_dam_unindent(tree node, int indent)
{
  dec_indent();
  dec_indent();
  dec_indent();
}

int
dump_dam(tree node, int indent)
{
  if (TREE_CODE(node) == FOR_STMT)
  {
    spit_dependence(node);
  }
}

int
dump_dam_expr(tree node, int indent)
{
  char  str[10];
  int   lineno;

  if (TREE_CODE(node) == EXPR_STMT)
  {
    dec_indent();
    dec_indent();
    dec_indent();
    if ((int) TREE_O_DEP(node) != 0)
    {
      lineno = (int) TREE_O_DEP(node);
      sprintf(str, "/*S%d*/", lineno);
      print_file_or_string(str);
    }
    else
      print_file_or_string("      ");
  }
}


int
dump_stmt_connect(tree node, int indent)
{
  /* dump info available only for stmts in heap related functions */
  if ((TREE_CODE(node) != TREE_LIST) && (HEAP_STUFF(dump_cur_func)) &&
    (HEAP_FLAG(dump_cur_func)))
  {
    heap_print_info_html(connect_dump_file, node, indent, dump_cur_func);
  }
}

int
dump_stmt_indirect_connect(tree node, int indent)
{
  /* dump info available only for stmts in heap related functions */
  if ((TREE_CODE(node) != TREE_LIST) && (HEAP_STUFF(dump_cur_func)) &&
    (HEAP_FLAG(dump_cur_func)))
  {
   if (TREE_CODE(node) == EXPR_STMT)
  {
    if ((TREE_CODE(STMT_BODY(node)) == MODIFY_EXPR) ||
      (TREE_CODE(STMT_BODY(node)) == CALL_EXPR))
    {
    heap_indirect_print_info_html(connect_indirect_dump_file, node, indent, 
                             dump_cur_func);
    }
  }

  }
}

int
dump_stmt_rw_connect(tree node, int indent)
{
  /* dump info available only for stmts in heap related functions */
  if ((TREE_CODE(node) != TREE_LIST) && (HEAP_STUFF(dump_cur_func)) &&
    (HEAP_FLAG(dump_cur_func)))
  {
#if 0
    spit_indent_string("");
    heap_rw_print_info(HEAP_RW_MATRIX(node), indent, dump_cur_func);
#endif
    heap_rw_print_info_html(heap_rw_dump_file, node, dump_cur_func);
  }
}

/* Rakesh, Oct 8/96: indicate loop invar access */
int
dump_heap_loop_invar(tree node, int indent)
{
  /* dump info available only for stmts in heap related functions */
  if ((node == NULL) || (!HEAP_STUFF(dump_cur_func)) ||
    (!HEAP_FLAG(dump_cur_func)))
    return;
  if ((TREE_CODE(node) == EXPR_STMT))
  {
    if (HEAP_LOOP_INVAR_FLAG(node))
    {
      spit_indent_string("/* loop invar access */");
      spit_string("\n");
    }
    if (HEAP_LOC_INVAR_FLAG(node))
    {
      spit_indent_string("/* loc invar access */");
      spit_string("\n");
    }
  }
}

/* Rakesh, Oct 24/96: indicate cse transform */
int
dump_heap_cse(tree node, int indent)
{
  /* dump info available only for stmts in heap related functions */
  if ((node == NULL) || (!HEAP_STUFF(dump_cur_func)) ||
    (!HEAP_FLAG(dump_cur_func)))
    return;
  if ((TREE_CODE(node) == EXPR_STMT) && (HEAP_CSE_FLAG(node)))
  {
    spit_indent_string("/* heap cse transform */");
    spit_string("\n");
  }
}

/** Points to dump using the Equivalence class analysis **/

int
dump_stmt_eqc(tree node, int indent)
{

  /*  only for indirect refs in the program */
  if (TREE_CODE(node) == EXPR_STMT){
    if ((TREE_CODE(STMT_BODY(node)) == MODIFY_EXPR) ||
        (TREE_CODE(STMT_BODY(node)) == CALL_EXPR))      {
      spit_indent_string("");
      eqc_print_info(STMT_BODY(node), indent);
    }
  }
}

/* Rakesh, Jan 23/1997: print fields accessed wrt heap pointers */
int
dump_heap_print_fields(tree node, int indent)
{
  /* dump info available only for stmts in heap related functions */
  if ((TREE_CODE(node) != TREE_LIST) && (HEAP_STUFF(dump_cur_func)) &&
    (HEAP_FLAG(dump_cur_func)))
  heap_fields_print_info_html(fields_dump_file, node, dump_cur_func);
}

